#include "runmap.h"
#include "osscan.h"

extern char *optarg;
extern int optind;
struct ops o;
extern char **environ;

int main(int argc, char *argv[], char *envp[]) {
  char *interactive_names[] = INTERACTIVE_NAMES;
  int numinames = sizeof(interactive_names) / sizeof(char *);
  int nameidx;
  char *nmapcalledas;
  char command[2048];
  int myargc, fakeargc;
  char **myargv = NULL, **fakeargv = NULL;
  char *cptr;
  int ret;
  int i;
  char runmapargs[1024];
  char fakeargs[1024];
  char runmappath[MAXPATHLEN];
  char *pptr;
  char path[4096];
  struct stat st;
  char *endptr;
  int interactivemode = 0;
  int fd;
  struct timeval tv;

  gettimeofday(&tv, NULL);
  srand((tv.tv_sec ^ tv.tv_usec) ^ getpid());

  options_init();

  signal(SIGINT, sigdie);
  signal(SIGTERM, sigdie);
  signal(SIGHUP, sigdie);

  signal(SIGCHLD, reaper);

  nmapcalledas = strrchr(argv[0], '/');
  if (!nmapcalledas) {
    nmapcalledas = argv[0];
  } else nmapcalledas++;

  if ((cptr = getenv("RUNMAP_ARGS"))) {
    snprintf(command, sizeof(command), "runmap %s", cptr);
    myargc = arg_parse(command, &myargv);
    if (myargc < 1) {
      fatal(" RUNMAP_ARG  .");
    }
    options_init();
    ret = nmap_main(myargc, myargv);
    arg_parse_free(myargv);
    return ret;
  }

  for(nameidx = 0; nameidx < numinames; nameidx++) {
    if (strcasecmp(nmapcalledas, interactive_names[nameidx]) == 0) {
      printf("   : argv[0] == %s\n", nmapcalledas);
      interactivemode = 1;
      break;
    }
  }

  if (interactivemode == 0 &&
      argc == 2 && strcmp("--interactive", argv[1]) == 0) {
    interactivemode = 1;
  }

  if (!interactivemode) {
    options_init();
    if (argc == 3 && strcmp("--resume", argv[1]) == 0) {
      if (gather_logfile_resumption_state(argv[2], &myargc, &myargv) == -1) {
	fatal(" %s .    .", argv[2]);
      }
      return nmap_main(myargc, myargv);
    }
    return nmap_main(argc, argv);
  }

  printf("\n\n%s %s :  , alex@nmap.ru (%s)\n", NMAP_NAME, NMAP_VERSION, NMAP_URL);

  printf(" : nmap 2.53, fyodor@insecure.org (http:/www.insecure.org)\n");

  printf("    .    h <enter>\n\n");

  while(1) {
    printf("runmap> ");
    fflush(stdout);
    fgets(command, sizeof(command), stdin);
    myargc = arg_parse(command, &myargv);
    if (myargc < 1) {
      printf("  .    h <enter>\n");
      continue;
    }
    if (strcasecmp(myargv[0], "h") == 0 ||
	strcasecmp(myargv[0], "help") == 0) {
      printinteractiveusage();
      continue;
    } else if (strcasecmp(myargv[0], "x") == 0 ||
	       strcasecmp(myargv[0], "q") == 0 ||
	       strcasecmp(myargv[0], "e") == 0 ||
	       strcasecmp(myargv[0], ".") == 0 ||
	       strcasecmp(myargv[0], "exit") == 0 ||
	       strcasecmp(myargv[0], "quit") == 0) {
      printf(" .\n");
      exit(0);
    } else if (strcasecmp(myargv[0], "r") == 0 ||
	       strcasecmp(myargv[0], "runmap") == 0) {
      options_init();
      o.interactivemode = 1;
      nmap_main(myargc, myargv);
    } else if (*myargv[0] == '!') {
      cptr = strchr(command, '!');
      system(cptr + 1);
    } else if (*myargv[0] == 'd') {
      o.debugging++;
    } else if (strcasecmp(myargv[0], "f") == 0) {
      switch((ret = fork())) {
      case 0:
	fakeargs[0] = runmappath[0] = '\0';
	strcpy(runmapargs, "RUNMAP_ARGS=");
	for(i=1; i < myargc; i++) {
	  if (strcasecmp(myargv[i], "--spoof") == 0) {
	    if (++i > myargc -1) {
	      fatal("  .\n");
	    }	
	    strncpy(fakeargs, myargv[i], sizeof(fakeargs));
	  } else if (strcasecmp(myargv[i], "--runmap_path") == 0) {
	    if (++i > myargc -1) {
	      fatal("  .\n");
	    }	
	    strncpy(runmappath, myargv[i], sizeof(runmappath));
	  } else {
	    if (strlen(runmapargs) + strlen(myargv[i]) + 1 < sizeof(runmapargs)) {
	      strcat(runmapargs, " ");
	      strcat(runmapargs, myargv[i]);
	    } else fatal("   .\n");
	  }	
	}
	if (o.debugging) {
	  error("%s   .\n", runmapargs);
	}
	if (putenv(runmapargs) == -1) {
	  pfatal("  RUNMAP_ARGS  .\n");
	}
	if (!*runmappath) {
	  if (stat(argv[0], &st) != -1 && !S_ISDIR(st.st_mode)) {
	    strncpy(runmappath, argv[0], sizeof(runmappath));
	  } else {
	    runmappath[0] = '\0';
	    if ((pptr = getenv("PATH"))) {
	      strncpy(path, pptr, sizeof(path));
	      pptr = path;
	      while(pptr && *pptr) {
		endptr = strchr(pptr, ':');
		if (endptr) {
		  *endptr = '\0';
		}
		snprintf(runmappath, sizeof(runmappath), "%s/%s", pptr, nmapcalledas);
		if (stat(runmappath, &st) != -1)
		  break;
		runmappath[0] = '\0';
		if (endptr) pptr = endptr + 1;
		else pptr = NULL;
	      }
	    }
	  }
	}
	if (!*runmappath) {
	  fatal("  RuNmap  .     ( --runmap_path).\n");
	}

	signal(SIGINT, SIG_DFL);
	signal(SIGTERM, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	signal(SIGSEGV, SIG_DFL);

	if (*fakeargs) {
	  fakeargc = arg_parse(fakeargs, &fakeargv);
	  if (fakeargc < 1) {
	    fatal("  --spoof  .\n");
	  }
	} else {
	  fakeargc = 1;
	  fakeargv = malloc(sizeof(char *) * 2);
	  fakeargv[0] = runmappath;
	  fakeargv[1] = NULL;
	}

	if (o.debugging) error("  %s", runmappath);
	if (!o.debugging) {	
	  fd = open("/dev/null", O_WRONLY);
	  if (fd != -1) {
	    dup2(fd, STDOUT_FILENO);
	    dup2(fd, STDERR_FILENO);
	  }
	}

	ret = execve(runmappath, fakeargv, environ);
	if (ret == -1) {
	  pfatal("  %s  .\n", runmappath);
	}
	break;
      case -1:
	gh_perror("   .\n");
	break;
      default:
	printf("[PID: %d]\n", ret);
	break;
      }
    } else {
      printf(" %s  .    h <enter>\n", myargv[0]);
      continue;
    }
    arg_parse_free(myargv);
  }
  return 0;

}

int nmap_main(int argc, char *argv[]) {
  char *p, *q;
  int i, j, arg, argvlen;
  FILE *inputfd = NULL;
  char *host_spec;
  short fastscan=0, randomize=1, resolve_all=0;
  short quashargv = 0;
  int numhosts_scanned = 0;
  char **host_exp_group;
  int num_host_exp_groups = 0;
  char *machinefilename = NULL, *kiddiefilename = NULL, *normalfilename = NULL;
  struct hostgroup_state hstate;
  int numhosts_up = 0;
  int starttime;
  unsigned short *ports = NULL;
  char myname[MAXHOSTNAMELEN + 1];
#if (defined(IN_ADDR_DEEPSTRUCT) || defined(SOLARIS))

  struct ftpinfo ftp = { FTPUSER, FTPPASS, "",  { { { 0 } } } , 21, 0};
#else
  struct ftpinfo ftp = { FTPUSER, FTPPASS, "", { 0 }, 21, 0};
#endif
  struct hostent *target = NULL;
  char **fakeargv;
  struct hoststruct *currenths;
  char emptystring[1];
  int sourceaddrwarning = 0;
  time_t timep;
  char mytime[128];
  int option_index;
  struct option long_options[] =
  {
    {"version", no_argument, 0, 'V'},
    {"verbose", no_argument, 0, 'v'},
    {"debug", optional_argument, 0, 'd'},
    {"help", no_argument, 0, 'h'},
    {"max_parallelism", required_argument, 0, 'M'},
    {"timing", required_argument, 0, 'T'},
    {"max_rtt_timeout", required_argument, 0, 0},
    {"min_rtt_timeout", required_argument, 0, 0},
    {"host_timeout", required_argument, 0, 0},
    {"scan_delay", required_argument, 0, 0},
    {"initial_rtt_timeout", required_argument, 0, 0},
    {"oN", required_argument, 0, 0},
    {"oM", required_argument, 0, 0},
    {"oS", required_argument, 0, 0},
    {"oH", required_argument, 0, 0},
    {"iL", required_argument, 0, 0},
    {"iR", no_argument, 0, 0},
    {"initial_rtt_timeout", required_argument, 0, 0},
    {"randomize_hosts", no_argument, 0, 0},
    {"rH", no_argument, 0, 0},
    {"vv", no_argument, 0, 0},
    {"append_output", no_argument, 0, 0},
    {"noninteractive", no_argument, 0, 0},
    {0, 0, 0, 0}
  };

#ifdef ROUTETHROUGHTEST
  {
    char *dev;
    struct in_addr dest;
    struct in_addr source;
    if (!resolve(argv[1], &dest))
      fatal("     %s\n", argv[1]);
      dev = routethrough(&dest, &source);
    if (dev)
      log_write(LOG_STDOUT, "   %s   %s  IP- %s\n", argv[1], dev, inet_ntoa(source));
    else log_write(LOG_STDOUT, "       %s.\n", argv[1]);

    exit(0);
  }
#endif


  fakeargv = (char **) safe_malloc(sizeof(char *) * (argc + 1));
  for(i=0; i < argc; i++) {
    fakeargv[i] = strdup(argv[i]);
  }
  fakeargv[argc] = NULL;

  emptystring[0] = '\0';

  if (argc < 2) printusage(argv[0], -1);


  optind = 1;
  while((arg = getopt_long_only(argc,fakeargv,"b:D:d::e:Ffg:hIi:M:m:NnOo:P:p:qRrS:s:T:Vv", long_options, &option_index)) != EOF) {
    switch(arg) {
    case 0:
      if (strcmp(long_options[option_index].name, "max_rtt_timeout") == 0) {
	o.max_rtt_timeout = atoi(optarg);
	if (o.max_rtt_timeout <= 5) {
	  fatal("  max_rtt_timeout     5 .\n");
	}
      } else if (strcmp(long_options[option_index].name, "min_rtt_timeout") == 0) {
	o.min_rtt_timeout = atoi(optarg);
	if (o.min_rtt_timeout > 50000) {
	  fatal(":    min_rtt_timeout  .\n");
	}
      } else if (strcmp(long_options[option_index].name, "host_timeout") == 0) {
	o.host_timeout = strtoul(optarg, NULL, 10);
	if (o.host_timeout <= 200) {
	  fatal("  host_timeout     200 .\n");
	}
      } else if (strcmp(long_options[option_index].name, "append_output") == 0) {
	o.append_output = 1;
      } else if (strcmp(long_options[option_index].name, "noninteractive") == 0) {
      } else if (strcmp(long_options[option_index].name, "scan_delay") == 0) {
	o.scan_delay = atoi(optarg);
	if (o.scan_delay <= 0) {
	  fatal("  scan_delay    0.\n");
	}
	o.max_parallelism = 1;
      } else if (strcmp(long_options[option_index].name, "randomize_hosts") == 0
		 || strcmp(long_options[option_index].name, "rH") == 0) {
	o.randomize_hosts = 1;
	o.host_group_sz = 2048;
      } else if (strcmp(long_options[option_index].name, "initial_rtt_timeout") == 0) {
	o.initial_rtt_timeout = atoi(optarg);
	if (o.initial_rtt_timeout <= 0) {
	  fatal("  scan_delay    0.\n");
	}
      } else if (strcmp(long_options[option_index].name, "oN") == 0) {
	normalfilename = optarg;
      } else if (strcmp(long_options[option_index].name, "oM") == 0) {
	machinefilename = optarg;
      } else if (strcmp(long_options[option_index].name, "oS") == 0) {
	kiddiefilename = optarg;
      } else if (strcmp(long_options[option_index].name, "oH") == 0) {
	fatal("    HTML     .\n");
      } else if (strcmp(long_options[option_index].name, "iL") == 0) {
	if (inputfd) {
	  fatal("        .\n");
	}
	if (!strcmp(optarg, "-")) {
	  inputfd = stdin;
	  log_write(LOG_STDOUT, "     stdin\n");
	} else {
	  inputfd = fopen(optarg, "r");
	  if (!inputfd) {
	    fatal("    %s  .\n", optarg);
	  }
	  log_write(LOG_STDOUT, "     : %s\n", optarg);
	}
      } else if (strcmp(long_options[option_index].name, "iR") == 0) {
	o.generate_random_ips = 1;
      } else if (strcmp(long_options[option_index].name, "vv") == 0) {
	o.verbose += 2;
      } else {
	fatal(" %s  .\n", long_options[option_index].name);
      }
      break;
    case 'b':
      o.bouncescan++;
      if (parse_bounce(&ftp, optarg) < 0) {
	fprintf(stderr, "\n  -b  .    (user:pass@server:port)\n       .\n    -h \n");
      }
      break;
    case 'D':
      p = optarg;
      do {
	q = strchr(p, ',');
	if (q) *q = '\0';
	if (!strcasecmp(p, "me")) {
	  if (o.decoyturn != -1)
	    fatal(" 'ME'         .\n");
	  o.decoyturn = o.numdecoys++;
	} else {
	  if (o.numdecoys >= MAX_DECOYS -1)
	    fatal("\n   %d  .\n  ,   MAX_DECOYS   runmap.h\n");
	  if (resolve(p, &o.decoys[o.numdecoys])) {
	    o.numdecoys++;
	  } else {
	    fatal("\n   : %s\n      IP-.\n", optarg);
	  }
	}
	if (q) {
	  *q = ',';
	  p = q+1;
	}
      } while(q);
      break;
    case 'd':
      if (optarg)
	o.debugging = o.verbose = atoi(optarg);
      else {
	o.debugging++; o.verbose++;
      }
      break;
    case 'e':
      strncpy(o.device, optarg,63); o.device[63] = '\0'; break;
    case 'F': fastscan++; break;
    case 'f': o.fragscan++; break;
    case 'g':
      o.magic_port = atoi(optarg);
      o.magic_port_set = 1;
      if (!o.magic_port) fatal(" -g    .\n");
      break;
    case 'h': printusage(argv[0], 0); break;
    case '?': printusage(argv[0], -1); break;
    case 'I': o.identscan++; break;
    case 'i':
      if (inputfd) {
	fatal("        .\n");
      }
      if (!strcmp(optarg, "-")) {
	inputfd = stdin;
	log_write(LOG_STDOUT, "     stdin\n");
      } else {
	inputfd = fopen(optarg, "r");
	if (!inputfd) {
	  fatal("   %s  .", optarg);
	}
	log_write(LOG_STDOUT, "     : %s\n", optarg);
      }
      break;
    case 'M':
      o.max_parallelism = atoi(optarg);
      if (o.max_parallelism < 1) fatal("  -M      1.\n");
      if (o.max_parallelism > MAX_SOCKETS_ALLOWED) {
	fprintf(stderr, "\n:      MAX_SOCKETS_ALLOWED   %d.\n  ,  #define    .\n", MAX_SOCKETS_ALLOWED);
	o.max_parallelism = MAX_SOCKETS_ALLOWED;
      }
      break;
    case 'm':
      machinefilename = optarg;
      break;
    case 'N': o.force++; break;
    case 'n': o.noresolve++; break;
    case 'O':
      o.osscan++;
      o.reference_FPs = parse_fingerprint_reference_file();
      break;
    case 'o':
      normalfilename = optarg;
      break;
    case 'P':
      if (*optarg == '\0' || *optarg == 'I')
	o.pingtype |= PINGTYPE_ICMP;
      else if (*optarg == '0' || *optarg == 'N' || *optarg == 'D')
	o.pingtype = PINGTYPE_NONE;
      else if (*optarg == 'S') {
	o.pingtype |= (PINGTYPE_TCP|PINGTYPE_TCP_USE_SYN);
	if (isdigit((int) *(optarg+1))) {
	  o.tcp_probe_port = atoi(optarg+1);
	  log_write(LOG_STDOUT, "TCP PING  %hu -   .\n", o.tcp_probe_port);
	} else if (o.verbose)
	  log_write(LOG_STDOUT, "TCP PING  %hu -   .\n", o.tcp_probe_port);
      }
      else if (*optarg == 'T' || *optarg == 'A') {
	o.pingtype |= (PINGTYPE_TCP|PINGTYPE_TCP_USE_ACK);
	if (isdigit((int) *(optarg+1))) {
	  o.tcp_probe_port = atoi(optarg+1);
	  log_write(LOG_STDOUT, "TCP PING  %hu -   .\n", o.tcp_probe_port);
	} else if (o.verbose)
	  log_write(LOG_STDOUT, "TCP PING  %hu -   .\n", o.tcp_probe_port);
      }
      else if (*optarg == 'B') {
	o.pingtype = (PINGTYPE_TCP|PINGTYPE_TCP_USE_ACK|PINGTYPE_ICMP);
	if (isdigit((int) *(optarg+1)))
	  o.tcp_probe_port = atoi(optarg+1);
	log_write(LOG_STDOUT, "TCP PING  %hu -   .\n", o.tcp_probe_port);
      }
      else {fatal("\n  -P  :   -P0, -PI, -PT, -PT80\n    ,     TCP PING.\n"); }
      break;
    case 'p':
      if (ports)
	fatal(" -p     ,    .\n");
      ports = getpts(optarg); break;
      if (!ports)
	fatal("   .\n");
    case 'q': quashargv++; break;
    case 'R': resolve_all++; break;
    case 'r':
      randomize = 0;
      break;
    case 'S':
      if (o.spoofsource)
	fatal("\n    IP-  !\n  -D <_1,[_2],[,ME],...>    IP-.\n");
      o.source = safe_malloc(sizeof(struct in_addr));
      o.spoofsource = 1;
      if (!resolve(optarg, o.source))
	fatal("   IP-  ,     IP-.\n");
      break;
    case 's':
      if (!*optarg) {
	fprintf(stderr, " -s    !\n: -sT (tcp-), -sS (SYN-), -sF (FIN-),\n-sU (UDP-)  -sP (Ping-).\n");
	printusage(argv[0], -1);
      }
      p = optarg;
      while(*p) {
	switch(*p) {
	case 'A': o.ackscan = 1; break;
	case 'B':  fatal(" 'B'  ,   '-b'.\n");
	  break;
	case 'F':  o.finscan = 1; break;
	case 'M':  o.maimonscan = 1; break;
	case 'N':  o.nullscan = 1; break;
	case 'P':  o.pingscan = 1; break;
	case 'R':  o.rpcscan = 1; break;
	case 'S':  o.synscan = 1; break;	
	case 'W':  o.windowscan = 1; break;
	case 'T':  o.connectscan = 1; break;
	case 'U':
	  o.udpscan++;
	  break;
	case 'X':  o.xmasscan++;break;
	default:  error("  %c   .\n",*p); printusage(argv[0], -1); break;
	}
	p++;
      }
      break;
    case 'T':
      if (*optarg == '0' || (strcasecmp(optarg, "Paranoid") == 0)) {
	o.max_parallelism = 1;
	o.scan_delay = 300000;
	o.initial_rtt_timeout = 300000;
      } else if (*optarg == '1' || (strcasecmp(optarg, "Sneaky") == 0)) {
	o.max_parallelism = 1;
	o.scan_delay = 15000;
	o.initial_rtt_timeout = 15000;
      } else if (*optarg == '2' || (strcasecmp(optarg, "Polite") == 0)) {
	o.max_parallelism = 1;
	o.scan_delay = 400;
      } else if (*optarg == '3' || (strcasecmp(optarg, "Normal") == 0)) {
      } else if (*optarg == '4' || (strcasecmp(optarg, "Aggressive") == 0)) {
	o.max_rtt_timeout = 1250;
	o.host_timeout = 300000;
	o.initial_rtt_timeout = 1000;
      } else if (*optarg == '5' || (strcasecmp(optarg, "Insane") == 0)) {
	o.max_rtt_timeout = 300;
	o.initial_rtt_timeout = 300;
	o.host_timeout = 75000;
      } else {
	fatal("\n  -T  .\n  \"Paranoid\", \"Sneaky\", \"Polite\", \"Normal\", \"Aggressive\", \"Insane\"\n    0 (Paranoid)  5 (Insane).\n");
      }
      break;
    case 'V':
      printf("\nRuNmap V. %s\n", NMAP_VERSION);
      exit(0);
      break;
    case 'v': o.verbose++; break;
    }
  }

  if (!o.debugging)
    signal(SIGSEGV, sigdie);

  if (!o.interactivemode)
    log_write(LOG_STDOUT|LOG_SKID, "\n%s %s  , alex@nmap.ru\n%s\n : nmap 2.53 fyodor@insecure.org\nhttp://www.insecure.org/nmap\n\n", NMAP_NAME, NMAP_VERSION, NMAP_URL);

  if (o.pingtype == PINGTYPE_UNKNOWN) {
    if (o.isr00t) o.pingtype = PINGTYPE_TCP|PINGTYPE_TCP_USE_ACK|PINGTYPE_ICMP;
    else o.pingtype = PINGTYPE_TCP;
  }


  if (normalfilename)
    log_open(LOG_NORMAL, o.append_output, normalfilename);
  if (machinefilename)
    log_open(LOG_MACHINE, o.append_output, machinefilename);
  if (kiddiefilename)
    log_open(LOG_SKID, o.append_output, kiddiefilename);


  if (!o.connectscan && !o.udpscan && !o.synscan && !o.windowscan && !o.finscan && !o.maimonscan &&  !o.nullscan && !o.xmasscan && !o.ackscan && !o.bouncescan && !o.pingscan) {
    o.connectscan++;
    if (o.verbose) error("\n      (UDP, TCP, ICMP).    tcp connect().\n -sP         ping-.\n");
  }

  if (o.pingtype != PINGTYPE_NONE && o.spoofsource) {
    error("\n:    -S     IP-,\n   -e <>  -P0.\n    IP-,    .\n");
  }

  if (o.connectscan && o.spoofsource) {
    error("\n:   -S      connect-.\n -sS    ,     IP-.\n");
  }

  if (fastscan && ports) {
    fatal("    -F  -p,    .\n");
  } else if (fastscan) {
    ports = getfastports(o.windowscan|o.synscan|o.connectscan|o.fragscan|o.finscan|o.maimonscan|o.bouncescan|o.nullscan|o.xmasscan|o.ackscan,o.udpscan);
  }

  if (o.pingscan && ports) {
    fatal("   -F  -p   PING-.\n");
  }

  if (o.pingscan && fastscan) {
    fatal("  (-F)   ping-.\n");
  }

  if (!ports) {
    ports = getdefaultports(o.windowscan|o.synscan|o.connectscan|o.fragscan|o.finscan|
			    o.maimonscan|o.bouncescan|o.nullscan|o.xmasscan|o.ackscan,
			    o.udpscan);
  }


  if (!o.tcp_probe_port) o.tcp_probe_port = 80;


  if (o.pingscan && (o.connectscan || o.udpscan || o.windowscan || o.synscan || o.finscan || o.maimonscan ||  o.nullscan || o.xmasscan || o.ackscan || o.bouncescan)) {
    fatal("Ping-      (    ping).\n");
  }

  if (!o.isr00t) {

    if (o.pingtype & PINGTYPE_ICMP) {
      error(":      ROOT -  ICMP Ping  TCP Ping.\n");
      o.pingtype = PINGTYPE_TCP;
    }

    if (o.finscan || o.windowscan || o.synscan || o.maimonscan || o.nullscan || o.xmasscan || o.ackscan
	|| o.udpscan) {
      fatal("  ,    ROOT.\n");
    }

    if (o.numdecoys > 0) {
      fatal("   IP-   ROOT.\n");
    }

    if (o.fragscan) {
      fatal("      ROOT.\n");
    }

    if (o.osscan) {
      fatal("      ROOT.\n");
    }
  }

  if (o.numdecoys > 0 && o.rpcscan) {
    error(": RPC-        IP-.\n");
  }
  if (o.bouncescan && o.pingtype != PINGTYPE_NONE)
    log_write(LOG_STDOUT, "\n:   '  FTP'    ,\n    -P0  ,     .\n");

  if (o.connectscan + o.windowscan + o.synscan + o.finscan + o.maimonscan + o.xmasscan + o.nullscan + o.ackscan  > 1) {
    fatal("\n    TCP-.\n     \n  : -sT, -sS, -sF, -sM, -sX, -sA, -sW,  -sN\n");
  }

  if (o.numdecoys > 0 && (o.bouncescan || o.connectscan)) {
    fatal("      connect-  '  FTP'\n");
  }

  if (o.fragscan && (o.connectscan ||
		     (o.udpscan && (o.windowscan + o.synscan + o.finscan + o.maimonscan +
				    o.xmasscan + o.ackscan + o.nullscan == 0))))
    fatal("IP-      SYN, FIN, NULL, XMAS, Maimon  ACK .\n");

  if (o.identscan && !o.connectscan) {
    error("Ident-     connect- (-sT) ...  .\n");
  }

  if (o.osscan && o.bouncescan)
    error(" '  FTP'    .\n");

#if !defined(LINUX) && !defined(OPENBSD) && !defined(FREEBSD) && !defined(NETBSD)
  if (o.fragscan) {
    fprintf(stderr, ":       ,   Linux, OpenBSD, FreeBSD,  NetBSD.    .\n");
  }
#endif

  if (o.max_parallelism > MAX_SOCKETS_ALLOWED) {
    error("\n:       MAX_SOCKETS_ALLOWED (%d).\n  ,  #define    .\n", MAX_SOCKETS_ALLOWED);
    o.max_parallelism = MAX_SOCKETS_ALLOWED;
  }

  if (o.osscan && o.pingscan) {
    fatal("\n:     ,   ping-.\n   -  ,  -sS, -sT, -sF\n");
  }

  if (o.magic_port_set && o.connectscan) {
    error("\n:   -g    connect() (-sT).\n   ,  -sS,    -.\n");
  }


  if (o.decoyturn == -1) {
    o.decoyturn = (o.numdecoys == 0)?  0 : get_random_uint() % o.numdecoys;
    o.numdecoys++;
    for(i=o.numdecoys-1; i > o.decoyturn; i--)
      o.decoys[i] = o.decoys[i-1];
  }


  if (o.source && !*o.device) {
    if (ipaddr2devname(o.device, o.source) != 0) {
      fatal("\n  IP-       !\n   IP- -   ,   -e eth0  -e ppp0   .\n");
    }
  }

  if (*o.device && !o.source) {
    o.source = safe_malloc(sizeof(struct in_addr));
    if (devname2ipaddr(o.device, o.source) == -1) {
      fatal("  IP-     %s.   .\n", o.device);
    }
  }



  if (o.bouncescan) {
    if (!inet_aton(ftp.server_name, &ftp.server)) {
      if ((target = gethostbyname(ftp.server_name)))
	memcpy(&ftp.server, target->h_addr_list[0], 4);
      else {
	fprintf(stderr, "     proxy-: %s\n",
		ftp.server_name);
	exit(1);
      }
    }  else if (o.verbose)
      log_write(LOG_STDOUT, "   FTP proxy- %s (%s).\n",
	      ftp.server_name, inet_ntoa(ftp.server));
  }
  fflush(stdout);

    timep = time(NULL);


    Strncpy(mytime, ctime(&timep), sizeof(mytime));
    chomp(mytime);
  log_write(LOG_NORMAL|LOG_MACHINE, "RuNmap (%s %s)  : ", NMAP_NAME, NMAP_VERSION, mytime);

  for(i=0; i < argc; i++) log_write(LOG_NORMAL|LOG_MACHINE,"%s ", fakeargv[i]);
  log_write(LOG_NORMAL|LOG_MACHINE,"\n");


  if (o.verbose)
    output_ports_to_machine_parseable_output(ports, o.numports, o.windowscan|o.synscan|o.connectscan|o.fragscan|o.finscan|o.maimonscan|o.bouncescan|o.nullscan|o.xmasscan|o.ackscan,o.udpscan);


  if (quashargv) {
    argvlen = strlen(argv[0]);
    if (argvlen < strlen(FAKE_ARGV))
      fatal("\n ,  RuNmap   ,   RuNmap,   .\n       RuNmap  superpuperportscannerandosdetector\n");
    strncpy(argv[0], FAKE_ARGV, strlen(FAKE_ARGV));
    for(i = strlen(FAKE_ARGV); i < argvlen; i++) argv[0][i] = '\0';
    for(i=1; i < argc; i++) {
      argvlen = strlen(argv[i]);
      for(j=0; j <= argvlen; j++)
	argv[i][j] = '\0';
    }
  }

  signal(SIGPIPE, SIG_IGN);
  if (o.max_parallelism && (i = max_sd()) && i < o.max_parallelism) {
    fprintf(stderr, ":   max_parallel_sockets   %d,      %d.   .\n", o.max_parallelism, i);
  }

  if (o.debugging > 1) log_write(LOG_STDOUT, "     : %d\n", o.max_parallelism);


  if (randomize)
    shortfry(ports, o.numports);

  starttime = time(NULL);


  host_exp_group = safe_malloc(o.host_group_sz * sizeof(char *));

  while(1) {
    while(num_host_exp_groups < o.host_group_sz &&
	  (host_spec = grab_next_host_spec(inputfd, argc, fakeargv))) {
      host_exp_group[num_host_exp_groups++] = strdup(host_spec);
    }
    if (num_host_exp_groups == 0)
      break;
    hostgroup_state_init(&hstate, o.host_group_sz, o.randomize_hosts,
			 host_exp_group, num_host_exp_groups);

    while((currenths = nexthost(&hstate)) && currenths->host.s_addr) {
      numhosts_scanned++;
      if (currenths->flags & HOST_UP)
	numhosts_up++;


      currenths->timedout = 0;
      if (o.host_timeout) {
	gettimeofday(&currenths->host_timeout, NULL);


	currenths->host_timeout.tv_sec += o.host_timeout / 1000;
	currenths->host_timeout.tv_usec += (o.host_timeout % 1000) * 1000;
	currenths->host_timeout.tv_sec += currenths->host_timeout.tv_usec / 1000000;
	currenths->host_timeout.tv_usec %= 1000000;
      }

      target = NULL;
      if (((currenths->flags & HOST_UP) || resolve_all) && !o.noresolve)
	target = gethostbyaddr((char *) &currenths->host, 4, AF_INET);
      if (target && *target->h_name) {
	currenths->name = strdup(target->h_name);
      }
      else {
	currenths->name = emptystring;
      }

      if (o.source) memcpy(&currenths->source_ip, o.source, sizeof(struct in_addr));
      if (!o.pingscan) {
	if (o.pingtype != PINGTYPE_NONE && o.verbose &&
	    !currenths->wierd_responses) {
	  if (currenths->flags & HOST_UP)
	    log_write(LOG_STDOUT, " %s (%s) .  .\n", currenths->name, inet_ntoa(currenths->host));
	  else  {
	    if (resolve_all) {	
	      log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT," %s (%s) ,    .\n", currenths->name, inet_ntoa(currenths->host));
	    }
	    else {
	      log_write(LOG_STDOUT," %s (%s) ,    .\n", currenths->name, inet_ntoa(currenths->host));
	    }
	    log_write(LOG_MACHINE, ": %s (%s)\t: \n", inet_ntoa(currenths->host), currenths->name);
	  }
	}
      }
      else {
	if (!currenths->wierd_responses) {	
	  if (currenths->flags & HOST_UP) {
	    log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT," %s (%s) .\n", currenths->name, inet_ntoa(currenths->host));
	    log_write(LOG_MACHINE,": %s (%s)\t: .\n", inet_ntoa(currenths->host), currenths->name);
	  }
	  else
	    if (o.verbose || o.debugging || resolve_all) {
	      if (resolve_all)
		log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT," %s (%s) .\n", currenths->name, inet_ntoa(currenths->host));
	      else log_write(LOG_STDOUT," %s (%s) .\n", currenths->name, inet_ntoa(currenths->host));
	      log_write(LOG_MACHINE, ": %s (%s)\t: .\n", inet_ntoa(currenths->host), currenths->name);
	    }
	}
      }

      if (currenths->wierd_responses) {
	if (!(currenths->flags & HOST_UP))
	  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n   %s (%s)    \n( %d    ping).    .\n",  currenths->name, inet_ntoa(currenths->host), currenths->wierd_responses);
	else
	  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n   %s (%s)    \n( %d    ping).    (    IP- ).\n",  currenths->name, inet_ntoa(currenths->host), currenths->wierd_responses);
	log_write(LOG_MACHINE,": %s (%s)\t: SMURF ( %d ).   .\n",  inet_ntoa(currenths->host), currenths->name, currenths->wierd_responses);
      }



      if (currenths->flags & HOST_UP &&
	  !o.pingscan) {

	if (currenths->flags & HOST_UP && !currenths->source_ip.s_addr && (o.windowscan || o.synscan || o.finscan || o.maimonscan || o.udpscan || o.nullscan || o.xmasscan || o.ackscan)) {
	  if (gethostname(myname, MAXHOSTNAMELEN) ||
	      !(target = gethostbyname(myname)))
	    fatal("   !    -S <IP- >  -e < >\n");
	  memcpy(&currenths->source_ip, target->h_addr_list[0], sizeof(struct in_addr));
	  if (! sourceaddrwarning) {
	    fprintf(stderr, "\n:      ,   %s .\n   ,   -S <IP- >.\n", inet_ntoa(currenths->source_ip));
	    sourceaddrwarning = 1;
	  }
	}


	if (!*currenths->device && currenths->flags & HOST_UP && (o.nullscan || o.xmasscan || o.ackscan || o.udpscan || o.finscan || o.maimonscan ||  o.synscan || o.osscan || o.windowscan) && (ipaddr2devname(currenths->device, &currenths->source_ip) != 0))
	  fatal("\n     .\n   -S (,     ).\n     SYN/FIN-   -S < IP>,\n  -e eth0 (  )     .\n");

	o.decoys[o.decoyturn] = currenths->source_ip;


	if (o.synscan) pos_scan(currenths, ports, SYN_SCAN);
	if (o.windowscan) pos_scan(currenths, ports, WINDOW_SCAN);
	if (o.connectscan) pos_scan(currenths, ports, CONNECT_SCAN);
	if (o.ackscan) pos_scan(currenths, ports, ACK_SCAN);

	if (o.finscan) super_scan(currenths, ports, FIN_SCAN);
	if (o.xmasscan) super_scan(currenths, ports, XMAS_SCAN);
	if (o.nullscan) super_scan(currenths, ports, NULL_SCAN);
	if (o.maimonscan) super_scan(currenths, ports, MAIMON_SCAN);
	if (o.udpscan) super_scan(currenths, ports, UDP_SCAN);

	if (o.bouncescan) {
	  if (ftp.sd <= 0) ftp_anon_connect(&ftp);
	  if (ftp.sd > 0) bounce_scan(currenths, ports, &ftp);
	}

	
	if (o.rpcscan)  pos_scan(currenths, ports, RPC_SCAN);


	if (o.osscan) {
	  os_scan(currenths, ports);
	}

	if (currenths->timedout) {
	  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,",     %s (%s) .\n", currenths->name,
		   inet_ntoa(currenths->host));
	  log_write(LOG_MACHINE,": %s (%s)\t:  \n",
			   inet_ntoa(currenths->host), currenths->name);
	}
	else if (!o.pingscan) {
	  assignignoredportstate(&currenths->ports);
	  printportoutput(currenths, &currenths->ports);
	  resetportlist(&currenths->ports);
	
	  if (o.osscan) {
	    if (currenths->seq.responses > 3) {
	      log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"%s", seqreport(&(currenths->seq)));
	      log_write(LOG_MACHINE,"\t : %d\n", currenths->seq.index);
	    }
	    if (currenths->FP_matches[0]) {
	      log_write(LOG_MACHINE,"\t: %s",  currenths->FP_matches[0]->OS_name);
	      i = 1;
	      while(currenths->FP_matches[i]) {
		log_write(LOG_MACHINE,"|%s", currenths->FP_matches[i]->OS_name);
		i++;
	      }
	      if (!currenths->FP_matches[1])
		log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n     : %s\n",
			 currenths->FP_matches[0]->OS_name);
	      else  {
		log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n    : %s\n",
			 currenths->FP_matches[0]->OS_name);
		i = 1;
		while(currenths->FP_matches[i]) {
		  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,", %s", currenths->FP_matches[i]->OS_name);
		  i++;
		}
	      }
	      log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n");	
	      if (currenths->goodFP >= 0 && (o.debugging || o.verbose > 1)) {
		log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT," :\n%s\n", fp2ascii(currenths->FPs[currenths->goodFP]));
	      }
	      log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n");
	    } else {
	      if (currenths->goodFP == ENOMATCHESATALL && o.scan_delay < 500) {
		log_write(LOG_NORMAL|LOG_SKID_NOXLT|LOG_STDOUT,"\n      .\n     ,\n http://www.insecure.org/cgi-bin/nmap-submit.cgi.\n TCP/IP:\n%s\n", mergeFPs(currenths->FPs, currenths->numFPs));
	      } else if (currenths->goodFP == ETOOMANYMATCHES) {
		log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"    .       .\n");
		if (o.debugging || o.verbose) {
		  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT," TCP/IP:\n%s\n",  mergeFPs(currenths->FPs, currenths->numFPs));
		}
	      }
	    }
	    for(i=0; i < currenths->numFPs; i++)
	      freeFingerPrint(currenths->FPs[i]);
	  }
	}

	if (o.debugging) log_write(LOG_STDOUT, "  :\n    (srtt): %d\n    (rttvar): %d\n  (to): %d\n\n", currenths->to.srtt, currenths->to.rttvar, currenths->to.timeout);
	log_write(LOG_MACHINE,"\n");
      }
      log_flush_all();
      if (*currenths->name) free(currenths->name);
    }

    for(i=0; i < num_host_exp_groups; i++)
      free(host_exp_group[i]);
    num_host_exp_groups = 0;
  }
  free(host_exp_group);
  timep = time(NULL);
  i = timep - starttime;
  if (numhosts_scanned == 1 && numhosts_up == 0)
    log_write(LOG_STDOUT, ":     .   ,   ,   ping-,   -P0\n");
  log_write(LOG_STDOUT|LOG_SKID, " RuNmap  -- %d %s ( %d %s)   %d %s\n\n", numhosts_scanned, (numhosts_scanned == 1)? "IP-" : "IP-", numhosts_up, (numhosts_up == 1)? "" : "",  i, (i == 1)? "": "");


  Strncpy(mytime, ctime(&timep), sizeof(mytime));
  chomp(mytime);

  log_write(LOG_NORMAL|LOG_MACHINE, "RuNmap   %s -- %d %s ( %d %s)   %d %s\n\n", mytime, numhosts_scanned, (numhosts_scanned == 1)? "IP-" : "IP-", numhosts_up, (numhosts_up == 1)? "" : "",  i, (i == 1)? "": "");



  for(i=0; i < argc; i++)
    free(fakeargv[i]);
  free(fakeargv);

  return 0;
}


int gather_logfile_resumption_state(char *fname, int *myargc, char ***myargv) {
  char *filestr;
  int filelen;
  char nmap_arg_buffer[1024];
  struct in_addr lastip;
  unsigned char *p, *q, *found;

  filestr = mmapfile(fname, &filelen, O_RDWR);
  if (!filestr) {
    fatal("    %s", fname);
  }

  if (filelen < 20) {
    fatal("  %s   --    .\n", fname);
  }


  filestr[filelen - 1] = '\0';


  p = strstr(filestr, " : ");
  p += 5;
  while(*p && !isspace(*p))
    p++;
  if (!*p) fatal("   %s.\n", fname);
  p++;
  if (*p == '\n' || !*p) fatal("   %s.\n", fname);

  q = strchr(p, '\n');
  if (!q || (q - p >= sizeof(nmap_arg_buffer) - 32))
    fatal("   %s.", fname);

  strcpy(nmap_arg_buffer, "runmap --append_output ");
  memcpy(nmap_arg_buffer + 21, p, q-p);
  nmap_arg_buffer[21 + q-p] = '\0';

  *myargc = arg_parse(nmap_arg_buffer, myargv);
  if (*myargc == -1) {
    fatal("   %s.", fname);
  }


  q = p;
  found = NULL;

  while((q = strstr(q, "\n: ")))
    found = q = q + 7;

  if (found) {
    q = strchr(found, ' ');
    if (!q) fatal("   %s.\n", fname);
    *q = '\0';
    if (inet_aton(found, &lastip) == 0)
      fatal("   %s.\n", fname);
    *q = ' ';
  } else {

    q = p;
    found = NULL;
    while((q = strstr(q, "\n   ")))
      found = q++;

    if (found) {
      found = strchr(found, '(');
      if (!found) fatal("   %s.\n", fname);
      found++;
      q = strchr(found, ')');
      if (!q) fatal("   %s.\n", fname);
      *q = '\0';
      if (inet_aton(found, &lastip) == 0)
	fatal("   %s.\n", found, fname);
      *q = ')';
    } else {
      error(":   --resume,   log-    .   .\n");
      lastip.s_addr = 0;
    }
  }
  o.resume_ip = lastip;


 filestr[filelen - 1] = '\n';
 munmap(filestr, filelen);
 return 0;
}

void options_init() {

  bzero((char *) &o, sizeof(struct ops));
  o.isr00t = !(geteuid());
  o.debugging = DEBUGGING;
  o.verbose = DEBUGGING;
  o.magic_port = 33000 + (get_random_uint() % 31000);
  o.pingtype = PINGTYPE_UNKNOWN;
  o.decoyturn = -1;
  o.nmap_stdout = stdout;
  o.host_group_sz = HOST_GROUP_SZ;
  o.min_rtt_timeout = MIN_RTT_TIMEOUT;
  o.max_rtt_timeout = MAX_RTT_TIMEOUT;
  o.initial_rtt_timeout = INITIAL_RTT_TIMEOUT;
  o.host_timeout = HOST_TIMEOUT;
  o.scan_delay = 0;

}

inline void max_rcvbuf(int sd) {
  int optval = 524288, optlen = sizeof(int);

  if (setsockopt(sd, SOL_SOCKET, SO_RCVBUF, (void *) &optval, optlen))
    if (o.debugging) perror("     .\n");
  if (o.debugging) {
    getsockopt(sd, SOL_SOCKET, SO_RCVBUF,(void *) &optval, &optlen);
    log_write(LOG_STDOUT, "   : %d\n", optval);
  }
}
int max_sd() {
  struct rlimit r;
  static int maxfds = -1;

  if (maxfds > 0)
    return maxfds;

#if(defined(RLIMIT_NOFILE))
  if (!getrlimit(RLIMIT_NOFILE, &r)) {
    r.rlim_cur = r.rlim_max;
    if (setrlimit(RLIMIT_NOFILE, &r))
      if (o.debugging) perror("   setrlimit (RLIMIT_NOFILE)\n");
    if (!getrlimit(RLIMIT_NOFILE, &r)) {
      maxfds =  MIN(r.rlim_cur, MAX_SOCKETS_ALLOWED);
      maxfds = MIN(maxfds, 250);
      return maxfds;
    } else return 0;
  }
#endif
#if(defined(RLIMIT_OFILE) && !defined(RLIMIT_NOFILE))
  if (!getrlimit(RLIMIT_OFILE, &r)) {
    r.rlim_cur = r.rlim_max;
    if (setrlimit(RLIMIT_OFILE, &r))
      if (o.debugging) perror("   setrlimit (RLIMIT_OFILE)\n");
    if (!getrlimit(RLIMIT_OFILE, &r)) {
      maxfds =  MIN(r.rlim_cur, MAX_SOCKETS_ALLOWED);
      maxfds = MIN(maxfds, 250);
      return maxfds;
    }
    else return 0;
  }
#endif
  return 0;
}

inline int block_socket(int sd) {
  int options;
  options = (~O_NONBLOCK) & fcntl(sd, F_GETFL);
  fcntl(sd, F_SETFL, options);
  return 1;
}

inline void broadcast_socket(int sd) {
  int one = 1;
  if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (void *)&one, sizeof(int)) != 0) {
    fprintf(stderr, " .\n");
    perror("setsockopt");
  }
}

inline void init_socket(int sd) {
  struct linger l;
  int res;
  static int bind_failed=0;
  struct sockaddr_in sin;

  l.l_onoff = 1;
  l.l_linger = 0;

  if (setsockopt(sd, SOL_SOCKET, SO_LINGER,  (void *) &l, sizeof(struct linger)))
    {
      fprintf(stderr, "  SO_LINGER  , : %d\n", errno);
      perror("setsockopt");
    }
  if (o.spoofsource && !bind_failed)
    {
      bzero(&sin,sizeof(sin));
      sin.sin_family=AF_INET;
      memcpy(&sin.sin_addr,o.source,sizeof(sin.sin_addr));
      res=bind(sd,(struct sockaddr*)&sin,sizeof(sin));
      if (res<0)
	{
	  fprintf(stderr, "init_socket:    (%s),  :%d\n", inet_ntoa(sin.sin_addr), errno);
	  perror("bind");
	  bind_failed=1;
	}
    }
}

unsigned short *getpts(char *origexpr) {
  unsigned char porttbl[65536];
  int portwarning = 0;
  long rangestart = -2343242, rangeend = -9324423;
  unsigned char *current_range;
  char *endptr;
  int i;
  unsigned short *ports;

  bzero(porttbl, sizeof(porttbl));
  o.numports = 0;

  current_range = origexpr;
  do {
    while(isspace(*current_range))
      current_range++;
    if (*current_range == '-') {
      rangestart = 1;
    }
    else if (isdigit(*current_range)) {
      rangestart = strtol(current_range, &endptr, 10);
      if (rangestart <= 0 || rangestart > 65535) {
	fatal("       1  65535 .\n");
      }
      current_range = endptr;
      while(isspace(*current_range)) current_range++;
    } else {
      fatal("  .  : \"-100,200-1024,3000-4000,60000-\"\n");
    }

    if (!*current_range || *current_range == ',') {
      rangeend = rangestart;
    } else if (*current_range == '-') {
      current_range++;
      if (!*current_range || *current_range == ',') {
	rangeend = 65535;
      } else if (isdigit(*current_range)) {
	rangeend = strtol(current_range, &endptr, 10);
	if (rangeend <= 0 || rangeend > 65535) {
	  fatal("       1  65535 .\n");
	}
	current_range = endptr;
      } else {
	fatal("  .  : \"-100,200-1024,3000-4000,60000-\"\n");
      }
    }


    while(rangestart <= rangeend) {
      if (porttbl[rangestart]) {
	if (!portwarning) {
	  error(":      .\n");
	  portwarning++;
	}
      } else o.numports++;
      porttbl[rangestart] = 1;
      rangestart++;
    }


    while(isspace(*current_range)) current_range++;
    if (*current_range && *current_range != ',') {
      fatal("  .  : \"-100,200-1024,3000-4000,60000-\"\n");
    }
    if (*current_range == ',')
      current_range++;
  } while(current_range && *current_range);

  if (o.numports == 0)
    fatal("  .     ,   'sP'.\n");

  ports = safe_malloc(sizeof(unsigned short) * (o.numports + 1));
  bzero(ports, sizeof(unsigned short) * (o.numports + 1));

  for(i=0, rangestart = 1; i < o.numports ; rangestart++) {
    assert(rangestart <= 65535);
    if (porttbl[rangestart]) {
      ports[i++] = rangestart;
    }
  }

  ports[o.numports] = 0;
  return ports;
}

void printusage(char *name, int rc) {
  printf(
	"\nRuNmap %s : runmap [() ] [] <   >\n"
	"    ('*' -   root)\n"
 	"-sT  TCP-  connect() ( )\n"
	"* -sS  TCP-  SYN ( )\n"
	"* -sU  UDP-\n"
	"  -sP Ping-  (  )\n"
	"* -sF,-sX,-sN  TCP-  FIN, Xmas  Null (  )\n"
	"  -sR/-I RPC/Identd- (    )\n"
	"   ( ,   ):\n"
	"* -O     \n"
	"  -p <>   . : '1-1024,1080,6666,31337'\n"
	"  -F   ,   nmap-services\n"
	"  -v .  .      \n"
	"  -P0  ping-  (   www.microsoft.com  .)\n"
	"* -D _1,_2[,...]     \n"
	"  -T <Paranoid|Sneaky|Polite|Normal|Aggressive|Insane>   \n"
	"  -n/-R /   DNS [ :  ]\n"
	"  -oN/-oM <_>    /    <_>\n"
	"  -iL <_>    .  '-'     stdin\n"
	"* -S <_IP>/-e <>      \n"
	"  --interactive     (     h)\n"
	" : runmap -v -sS -O www.my.com 192.168.0.0/16 '192.88-90.*.*'\n"
	"        .\n\n", NMAP_VERSION);
  exit(rc);
}

void printinteractiveusage() {
  printf(
"\n                           :\n
r <> --         .
                 (      ).
! <>   --   
x             --   
f [--spoof <>] [--nmap_path <>] <>--  RuNmap   
            (    ).      
              ( -oM  -oN).       --spoof,
            RuNmap    ps-.    
             RuNmap,   --runmap_path.
r -h          --    RuNmap
h             --   .
\n                                    :\n
r -sS -O -v example.com/24
f --spoof \"/usr/local/bin/pico -z hello.c\" -sS -oN /tmp/e.log example.com/24\n\n");
}

char *seqreport(struct seq_info *seq) {
  static char report[512];
  char tmp[256];
  char *p;
  int i;

  snprintf(report, sizeof(report), "  ISN: %s\n   : %d (%s)\n", seqclass2ascii(seq->class), seq->index, (seq->index < 10)? " " : (seq->index < 80)? "" : (seq->index < 3000)? "" : (seq->index < 5000)? "" : (seq->index < 100000)? " " : "");
  if (o.verbose) {
    tmp[0] = '\n';
    tmp[1] = '\0';
    p = tmp + 1;
    strcpy(p, " ISN:      ");
    p += 18;
    for(i=0; i < seq->responses; i++) {
      p += snprintf(p, 16, "%X ", seq->seqs[i]);
    }
    *--p = '\n';
    strcat(report, tmp);
  }
  return report;
}

char *seqclass2ascii(int class) {
  switch(class) {
  case SEQ_CONSTANT:
    return " ";
  case SEQ_64K:
    return " 64K";
  case SEQ_TD:
    return " - ";
  case SEQ_i800:
    return "  800";
  case SEQ_RI:
    return "  ";
  case SEQ_TR:
    return "  ";
  case SEQ_UNKNOWN:
    return " ";
  default:
    return " ";
  }
}

char *grab_next_host_spec(FILE *inputfd, int argc, char **fakeargv) {
  static char host_spec[512];
  int host_spec_index;
  int ch;
  unsigned char *ipc;
  struct in_addr ip;

  if (o.generate_random_ips) {
    do {
      ipc = (unsigned char *) &ip.s_addr;
      get_random_bytes(ipc, 4);
    } while(ipc[0] == 10 || ipc[0] > 224 || ipc[0] == 127 || !ipc[0]);
    strcpy(host_spec, inet_ntoa(ip));
  } else if (!inputfd) {
    return((optind < argc)?  fakeargv[optind++] : NULL);
  } else {
    host_spec_index = 0;
    while((ch = getc(inputfd)) != EOF) {
      if (ch == ' ' || ch == '\n' || ch == '\t' || ch == '\0') {
	if (host_spec_index == 0) continue;
	host_spec[host_spec_index] = '\0';
	return host_spec;
      } else if (host_spec_index < 511) {
	host_spec[host_spec_index++] = (char) ch;
      } else fatal("        (> %d ).\n", sizeof(host_spec));
    }
    host_spec[host_spec_index] = '\0';
  }
  if (!*host_spec) return NULL;
  return host_spec;
}

char *statenum2str(int state) {
  switch(state) {
  case PORT_OPEN: return ""; break;
  case PORT_FIREWALLED: return ""; break;
  case PORT_UNFIREWALLED: return ""; break;
  case PORT_CLOSED: return ""; break;
  default: return " "; break;
  }
  return " ";
}


void printportoutput(struct hoststruct *currenths, portlist *plist) {
  char protocol[4];
  char rpcinfo[64];
  char rpcmachineinfo[64];
  char portinfo[64];
  char *state;
  char serviceinfo[64];
  char *name;
  int first = 1;
  struct servent *service;
  struct port *current;
  int numignoredports;
  int portno, protocount;
  struct port **protoarrays[2];

  numignoredports = plist->state_counts[plist->ignored_port_state];

  assert(numignoredports <= plist->numports);

  if (numignoredports == plist->numports) {
    log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,
              "%s %d %s    %s (%s) %s: %s\n",
	      (numignoredports == 1)? "" : "", numignoredports,
	      (numignoredports == 1)? "" : "", currenths->name,
	      inet_ntoa(currenths->host),
	      (numignoredports == 1)? " " : " ",
	      statenum2str(currenths->ports.ignored_port_state));
    log_write(LOG_MACHINE,": %s (%s)\t: ",
	      inet_ntoa(currenths->host), currenths->name);
    return;
  }

  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"   %s (%s):\n", currenths->name,
	    inet_ntoa(currenths->host));
  log_write(LOG_MACHINE,": %s (%s)", inet_ntoa(currenths->host),
	    currenths->name);

  if (numignoredports > 0) {
    log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"( %d %s, %s : %s)\n", numignoredports, (numignoredports == 1)? "" : "", (numignoredports == 1)? "      " : "      ", statenum2str(plist->ignored_port_state));
  }

  if (!o.rpcscan) {
    log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"              ");
  } else {
    log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"              (RPC)");
  }
  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"%s", (o.identscan)? ((o.rpcscan)? "            \n" : "                  \n") :"\n");
  log_write(LOG_MACHINE,"\t: ");

  protoarrays[0] = plist->tcp_ports;
  protoarrays[1] = plist->udp_ports;
  current = NULL;
  for(portno = 1; portno < 65536; portno++) {
    for(protocount = 0; protocount < 2; protocount++) {
      if (protoarrays[protocount] && protoarrays[protocount][portno])
	current = protoarrays[protocount][portno];
      else continue;

      if (current->state != plist->ignored_port_state) {
	if (!first) log_write(LOG_MACHINE,", ");
	else first = 0;
	strcpy(protocol,(current->proto == IPPROTO_TCP)? "tcp": "udp");
	snprintf(portinfo, sizeof(portinfo), "%d/%s", current->portno, protocol);
	state = statenum2str(current->state);
	service = nmap_getservbyport(htons(current->portno), protocol);
	
	if (o.rpcscan) {
	  switch(current->rpc_status) {
	  case RPC_STATUS_UNTESTED:
	    rpcinfo[0] = '\0';
	    strcpy(rpcmachineinfo, "");
	    break;
	  case RPC_STATUS_UNKNOWN:
	    strcpy(rpcinfo, "(RPC ( ))");
	    strcpy(rpcmachineinfo, "R");
	    break;
	  case RPC_STATUS_NOT_RPC:
	    rpcinfo[0] = '\0';
	    strcpy(rpcmachineinfo, "N");
	    break;
	  case RPC_STATUS_GOOD_PROG:
	    name = nmap_getrpcnamebynum(current->rpc_program);
	    snprintf(rpcmachineinfo, sizeof(rpcmachineinfo), "(%s:%li*%i-%i)", (name)? name : "", current->rpc_program, current->rpc_lowver, current->rpc_highver);
	    if (!name) {
	      snprintf(rpcinfo, sizeof(rpcinfo), "(#%li (unknown) V%i-%i)", current->rpc_program, current->rpc_lowver, current->rpc_highver);
	    } else {
	      if (current->rpc_lowver == current->rpc_highver) {
		snprintf(rpcinfo, sizeof(rpcinfo), "(%s V%i)", name, current->rpc_lowver);
	      } else
		snprintf(rpcinfo, sizeof(rpcinfo), "(%s V%i-%i)", name, current->rpc_lowver, current->rpc_highver);
	    }
	    break;
	  default:
	    fatal("  rpc %d", current->rpc_status);
	    break;
	  }
	  snprintf(serviceinfo, sizeof(serviceinfo), "%s%s%s", (service)? service->s_name : ((*rpcinfo)? "" : "unknown"), (service)? " " : "",  rpcinfo);
	} else {
	  Strncpy(serviceinfo, (service)? service->s_name : "unknown" , sizeof(serviceinfo));
	  strcpy(rpcmachineinfo, "");
	}
	log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"%-11s%-12s%-24s", portinfo, state, serviceinfo);
	log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"%s\n", (current->owner)? current->owner : "");
	
	log_write(LOG_MACHINE,"%d/%s/%s/%s/%s/%s//", current->portno, state,
		  protocol, (current->owner)? current->owner : "",
		  (service)? service->s_name: "", rpcmachineinfo);
	
      }
    }
  }
  log_write(LOG_NORMAL|LOG_SKID|LOG_STDOUT,"\n");
  log_write(LOG_MACHINE, "\t : %s (%d)\n", statenum2str(plist->ignored_port_state), plist->state_counts[plist->ignored_port_state]);
}

unsigned long calculate_sleep(struct in_addr target) {
  struct timeval begin, end;
  int sd;
  struct sockaddr_in sock;
  int res;

  if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
    {perror(" ."); exit(1);}

  sock.sin_family = AF_INET;
  sock.sin_addr.s_addr = target.s_addr;
  sock.sin_port = htons(o.magic_port);

  gettimeofday(&begin, NULL);
  if ((res = connect(sd, (struct sockaddr *) &sock,
		     sizeof(struct sockaddr_in))) != -1)
    fprintf(stderr, "\n:       -g (  o.magic_port),\n     !\n");
  close(sd);
  gettimeofday(&end, NULL);
  if (end.tv_sec - begin.tv_sec > 5)
    return 0;
  return (end.tv_sec - begin.tv_sec) * 1000000 + (end.tv_usec - begin.tv_usec);
}

int check_ident_port(struct in_addr target) {
  int sd;
  char buf[4096];
  struct sockaddr_in sock;
  int res;
  struct sockaddr_in stranger;
  int sockaddr_in_len = sizeof(struct sockaddr_in);
  fd_set fds_read, fds_write;
  struct timeval tv;
  tv.tv_sec = o.initial_rtt_timeout / 1000;
  tv.tv_usec = (o.initial_rtt_timeout % 1000) * 1000;
  if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
    {perror(" ."); exit(1);}
  unblock_socket(sd);
  sock.sin_family = AF_INET;
  sock.sin_addr.s_addr = target.s_addr;
  sock.sin_port = htons(113);
  FD_SET(sd, &fds_read);
  FD_SET(sd, &fds_write);
  res = connect(sd, (struct sockaddr *) &sock, sizeof(struct sockaddr_in));
  if (res != -1)
    goto success;
  if (errno == ECONNREFUSED)
    goto failure;
  if ((res = select(sd+1, &fds_read, &fds_write, NULL, &tv)) > 0) {
    if (FD_ISSET(sd, &fds_read) && FD_ISSET(sd, &fds_write)) {
      res = recvfrom(sd, buf,4096, 0, (struct sockaddr *) & stranger, &sockaddr_in_len);
      if (res >= 0) goto success;
      goto failure;
    }
    else if (FD_ISSET(sd, &fds_write)) {
      res = send(sd, buf, 0, 0);
      if (res < 0) goto failure;
      goto success;
    } else if (FD_ISSET(sd, &fds_read)) {
      fprintf(stderr, "\n         read-only.\n ,          (alex@nmap.ru).\n");
      goto success;
    }
    else {
      fprintf(stderr, "   !       (alex@nmap.ru).\n");
      goto failure;
    }
  }

 failure:
  close(sd);
  if (o.debugging || o.verbose) log_write(LOG_STDOUT, " identd .\n");
  return 0;

 success:
  close(sd);
  if (o.debugging || o.verbose) log_write(LOG_STDOUT, " identd .\n");
  return 1;
}


int getidentinfoz(struct in_addr target, int localport, int remoteport,
		  char *owner) {
  int sd;
  struct sockaddr_in sock;
  int res;
  char request[16];
  char response[1024];
  char *p,*q;
  char  *os;

  owner[0] = '\0';
  if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
    {perror(" ."); exit(1);}

  sock.sin_family = AF_INET;
  sock.sin_addr.s_addr = target.s_addr;
  sock.sin_port = htons(113);
  usleep(50000);
  res = connect(sd, (struct sockaddr *) &sock, sizeof(struct sockaddr_in));

  if (res < 0) {
    if (o.debugging)
      fprintf(stderr, " Identd .   ,  .\n");
    close(sd);
    return -1;
  }
  snprintf(request, sizeof(request), "%hu,%hu\r\n", remoteport, localport);
  if (o.debugging > 1) log_write(LOG_STDOUT, "  identd ,  : %s", request);
  if (write(sd, request, strlen(request) + 1) == -1) {
    perror(" identd");
    close(sd);
    return 0;
  }
  else if ((res = read(sd, response, 1024)) == -1) {
    perror("  identd");
    close(sd);
    return 0;
  }
  else {
    close(sd);
    if (o.debugging > 1) log_write(LOG_STDOUT, " %d   identd: %s\n", res, response);
    if ((p = strchr(response, ':'))) {
      p++;
      if ((q = strtok(p, " :"))) {
	if (!strcasecmp(q, "error")) {
	  if (strstr(response, "HIDDEN-USER") || strstr(response, "hidden-user")) {
	    log_write(LOG_STDOUT, "identd  HIDDEN-USER.  .\n");
	    return -1;
	  }
	  if (o.debugging) log_write(LOG_STDOUT, "Identd  ERROR   %d\n", remoteport);
	  return 0;
	}
	if ((os = strtok(NULL, " :"))) {
	  if ((p = strtok(NULL, " :"))) {
	    if ((q = strchr(p, '\r'))) *q = '\0';
	    if ((q = strchr(p, '\n'))) *q = '\0';
	    strncpy(owner, p, 512);
	    owner[512] = '\0';
	  }
	}
      }
    }
  }
  return 1;
}


void super_scan(struct hoststruct *target, unsigned short *portarray, stype scantype) {
  int initial_packet_width;
  int packet_incr = 4;
  double fallback_percent = 0.7;
  int rawsd;
  char myname[513];
  int scanflags = 0;

  int dropped = 0;
  int freshportstried = 0;
  int senddelay = 0;
  pcap_t *pd;
  int bytes;
  struct ip *ip, *ip2;
  struct tcphdr *tcp;
  struct bpf_program fcode;
  char err0r[PCAP_ERRBUF_SIZE];
  char filter[512];
  char *p;
  int changed = 0;
  int numqueries_outstanding = 0;
  double numqueries_ideal;
  int max_width;
  int tries = 0;
  int tmp = 0;
  unsigned int localnet, netmask;
  int starttime;
  unsigned short newport;
  int newstate = 999;
  struct hostent *myhostent = NULL;
  struct portinfo *scan, *openlist, *current, *testinglist, *next;
  int portlookup[65536];
  struct timeval now, end;
  int packcount, timedout;
  int UDPPacketWarning = 0;
  int i;
  unsigned short *data;
  int packet_trynum = 0;
  int windowdecrease = 0;
  struct icmp *icmp;
  int portno;
  struct port *current_port_tmp;

  if (target->timedout)
    return;

  if (o.debugging)
    log_write(LOG_STDOUT, " super_scan\n");

  max_width = (o.max_parallelism)? o.max_parallelism : 125;
  numqueries_ideal = initial_packet_width = MIN(max_width, 10);

  memset(portlookup, 255, 65536 * sizeof(int));
  scan = safe_malloc(o.numports * sizeof(struct portinfo));

  for(i = 0; i < o.numports; i++) {
    scan[i].state = PORT_FRESH;
    scan[i].portno = portarray[i];
    scan[i].trynum = 0;
    scan[i].prev = i-1;
    if (i < o.numports -1) scan[i].next = i+1;
    else scan[i].next = -1;
    portlookup[portarray[i]] = i;
  }

  current = testinglist = &scan[0];
  openlist = NULL;




  if ((rawsd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
    pfatal("    super_scan\n");
  broadcast_socket(rawsd);

  if (!target->source_ip.s_addr) {
    if (gethostname(myname, MAXHOSTNAMELEN) != 0 &&
	!((myhostent = gethostbyname(myname))))
      fatal("   !    -S <IP- >  -e < >\n");
    memcpy(&target->source_ip, myhostent->h_addr_list[0], sizeof(struct in_addr));
    if (o.debugging || o.verbose)
      log_write(LOG_STDOUT, "  IP-: %s\n",
	      inet_ntoa(target->source_ip));
  }


  pd = my_pcap_open_live(target->device, 92,  (o.spoofsource)? 1 : 0, 10);

  if (pcap_lookupnet(target->device, &localnet, &netmask, err0r) < 0)
    fatal("   : %s\n", err0r);
  p = strdup(inet_ntoa(target->host));
   snprintf(filter, sizeof(filter), "(icmp and dst host %s) or (tcp and src host %s and dst host %s and (dst port %d or dst port %d))", inet_ntoa(target->source_ip), p, inet_ntoa(target->source_ip), o.magic_port , o.magic_port + 1);
  free(p);
  if (islocalhost(&(target->host)))
    filter[0] = '\0';

  if (o.debugging)
    log_write(LOG_STDOUT, "   \n%s\n", filter);
  if (pcap_compile(pd, &fcode, filter, 0, netmask) < 0)
    fatal("    \n%s\n", pcap_geterr(pd));
  if (pcap_setfilter(pd, &fcode) < 0)
    fatal("    \n%s\n", pcap_geterr(pd));

  if (scantype == XMAS_SCAN) scanflags = TH_FIN|TH_URG|TH_PUSH;
  else if (scantype == NULL_SCAN) scanflags = 0;
  else if (scantype == FIN_SCAN) scanflags = TH_FIN;
  else if (scantype == MAIMON_SCAN) scanflags = TH_FIN|TH_ACK;
  else if (scantype != UDP_SCAN) { fatal("super_scan:   .\n"); }

  starttime = time(NULL);

  if (o.debugging || o.verbose)
    log_write(LOG_STDOUT, " FIN,NULL, UDP,  Xmas-  %s (%s)\n", target->name, inet_ntoa(target->host));


  do {
    changed = 0;
    if (tries > 3 && senddelay == 0) senddelay = 10000;
							
    while(testinglist != NULL)
      {
	gettimeofday(&now, NULL);

	if (o.host_timeout && numqueries_outstanding > 0 && (TIMEVAL_MSEC_SUBTRACT(now, target->host_timeout) >= 0))
	  {
	    target->timedout = 1;
	    goto superscan_timedout;
	  }

	for(current = testinglist; current ; current = next) {
	  next = (current->next > -1)? &scan[current->next] : NULL;
	  if (current->state == PORT_TESTING) {
	    if (TIMEVAL_SUBTRACT(now, current->sent[current->trynum]) > target->to.timeout) {
	      if (current->trynum > 0) {
		if (o.debugging > 1) { log_write(LOG_STDOUT, " %lu     .\n", current->portno); }
		freshportstried--;
		current->state = PORT_OPEN;
		if (current->next > -1) scan[current->next].prev = current->prev;
		if (current->prev > -1) scan[current->prev].next = current->next;
		if (current == testinglist)
		  testinglist = (current->next >= 0)?  &scan[current->next] : NULL;
		current->next = current->prev = -1;

		if (!openlist) openlist = current;
		else {
		  current->next = openlist - scan;
		  openlist = current;
		  scan[current->next].prev = current - scan;	
		}
		numqueries_outstanding--;
	      } else {
		if (o.scan_delay) enforce_scan_delay(NULL);
		if (o.debugging > 1) { log_write(LOG_STDOUT, "  ,    %lu\n", current->portno); }
		current->trynum++;

		i = (o.magic_port_set)? o.magic_port : o.magic_port + 1;
		gettimeofday(&current->sent[1], NULL);
		now = current->sent[1];
		if (o.fragscan)
		  send_small_fragz_decoys(rawsd, &target->host, 0,i, current->portno, scanflags);
		else if (scantype != UDP_SCAN)
		  send_tcp_raw_decoys(rawsd, &target->host, i,
				      current->portno, 0, 0, scanflags, 0, NULL, 0,
				      0, 0);
		else send_udp_raw_decoys(rawsd, &target->host, i,
					 current->portno, NULL ,0);	
		if (senddelay && scantype == UDP_SCAN) usleep(senddelay);
	      }
	    }
	  } else {
	    if (numqueries_outstanding >= (int) numqueries_ideal) break;
	    if (o.scan_delay) enforce_scan_delay(NULL);
	    if (o.debugging > 1) log_write(LOG_STDOUT, "     %lu\n", current->portno);
	    freshportstried++;
	    current->state = PORT_TESTING;
	    numqueries_outstanding++;
	    gettimeofday(&current->sent[0], NULL);
	    if (o.fragscan)
	      send_small_fragz_decoys(rawsd, &target->host, 0, o.magic_port, current->portno, scanflags);
	    else if (scantype != UDP_SCAN)
	      send_tcp_raw_decoys(rawsd, &target->host, o.magic_port,
				  current->portno, 0, 0, scanflags, 0, NULL, 0, 0, 0);
	    else send_udp_raw_decoys(rawsd, &target->host, o.magic_port,
				     current->portno, NULL, 0);
	    if (scantype == UDP_SCAN && senddelay) usleep(senddelay);
	  }
	}

	if (o.debugging > 1) log_write(LOG_STDOUT, "  : %d\n", (int) numqueries_ideal);
	tmp++;
	windowdecrease = 0;
	timedout = packcount = 0;
	gettimeofday(&now, NULL);
	if (o.host_timeout && (TIMEVAL_MSEC_SUBTRACT(now, target->host_timeout) >= 0))
	  {
	    target->timedout = 1;
	    goto superscan_timedout;
	  }
	while (!timedout && numqueries_outstanding > 0 && (ip = (struct ip*) readip_pcap(pd, &bytes, target->to.timeout)))
	  {
	    if (++packcount >= 30) {
	      gettimeofday(&end, NULL);
	      if (TIMEVAL_SUBTRACT(end, now) > 8000000)
		timedout = 1;
	    }
	    if (bytes < (4 * ip->ip_hl) + 4)
	      continue;	
	    current = NULL;
	    if (ip->ip_p == IPPROTO_ICMP ||
		ip->ip_src.s_addr == target->host.s_addr) {
	      if (ip->ip_p == IPPROTO_TCP) {
		tcp = (struct tcphdr *) (((char *) ip) + 4 * ip->ip_hl);
		if (tcp->th_flags & TH_RST) {	
		  newstate = PORT_CLOSED;
		  newport = ntohs(tcp->th_sport);
		  if (portlookup[newport] < 0) {
		    if (o.debugging) {
		      log_write(LOG_STDOUT, "      %d\n", ntohs(tcp->th_sport));
		      readtcppacket((char *)ip, bytes);
		    }
		    current = NULL;
		    continue;
		  }	
		  current = &scan[portlookup[newport]];
		
		  if (ntohs(tcp->th_dport) != o.magic_port &&
		      ntohs(tcp->th_dport) != o.magic_port + 1) {
		    if (o.debugging)  {		
		      error("  %d   TCP-   %d\n", newport, ntohs(tcp->th_dport));
		    }
		    continue;		
		  }
		
		  if (current->state != PORT_TESTING && o.debugging) {
		    error(" TCP-    %d   %d\n",
			  newport, current->state);
		  }
		
		  if (!o.magic_port_set) {
		    packet_trynum = ntohs(tcp->th_dport) - o.magic_port;
		    if ((packet_trynum|1) != 1) packet_trynum = -1;
		  }  else packet_trynum = -1;
		  if (current->trynum == 0) packet_trynum = 0;
		} else { continue; }
	
	      } else if (ip->ip_p == IPPROTO_ICMP) {
		icmp = (struct icmp *) ((char *)ip + 4 * ip->ip_hl);
		ip2 = (struct ip *) (((char *) icmp) + 8);
		if (ip2->ip_dst.s_addr != target->host.s_addr)
		  continue;
		data = (unsigned short *) ((char *)ip2 + 4 * ip2->ip_hl);
		if (icmp->icmp_type == 3) {
		  newport = ntohs(data[1]);
		  if (portlookup[newport] < 0) {
		    if (o.debugging) {
		      log_write(LOG_STDOUT, "  ICMP-:  3  %d   %d \n", icmp->icmp_code, newport);
		      readtcppacket((char *)ip, bytes);		
		    }
		    continue;		
		  }
		  current = &scan[portlookup[newport]];
		  if (!o.magic_port_set) {
		    packet_trynum = ntohs(data[0]) - o.magic_port;
		    if ((packet_trynum|1) != 1) packet_trynum = -1;
		  } else {
		    if (current->trynum == 0)  {
		      packet_trynum = 0;
		    }
		    else packet_trynum = -1;
		  }
		
		  switch(icmp->icmp_code) {
		
		  case 2:
		    newstate = PORT_FIREWALLED;
		    break;
		
		  case 3:		
		    if (scantype == UDP_SCAN) {
		      newstate = PORT_CLOSED;
		    } else newstate = PORT_FIREWALLED;
		    break;
		
		  case 9:
		  case 10:
		  case 13:
		    newstate = PORT_FIREWALLED;
		    break;		
		
		  default:
		    if (o.debugging) {
		      error(" ICMP-   - : %d\n", icmp->icmp_code);
		      hdump((unsigned char *)icmp, ntohs(ip->ip_len) -
			    sizeof(struct ip));
		    }
		    continue;
		  }
		}
	      } else if (ip->ip_p == IPPROTO_UDP) {
		if (UDPPacketWarning == 0) {
		  UDPPacketWarning = 1;
		  if (o.debugging)
		    error(" UDP-.\n");
		}
		continue;
	      }
	
	      if (current) {	
		if (current->state == PORT_CLOSED && (packet_trynum < 0)) {
		  target->to.rttvar *= 1.2;
		  if (o.debugging) { log_write(LOG_STDOUT, "\n   ,      .\n rttvar   %d \n", target->to.rttvar);
		  }
		}
		if (packet_trynum > -1) {		
		  adjust_timeouts(current->sent[packet_trynum], &(target->to));
		  numqueries_ideal = MIN(numqueries_ideal + (packet_incr/numqueries_ideal), max_width);
		  if (packet_trynum > 0 && current->trynum > 0) {
		    dropped++;
		    if (freshportstried > 50 && ((double) dropped/freshportstried) > 0.3) {
		      if (!senddelay) senddelay = 50000;
		      else senddelay = MIN(senddelay * 2, 1000000);
		      if (senddelay >= 200000 && scantype == UDP_SCAN)
			max_width = MIN(max_width,2);
		      freshportstried = 0;
		      dropped = 0;

		      if (o.verbose || o.debugging)
			log_write(LOG_STDOUT, "   .      %d \n", senddelay);
		    }
		    if (windowdecrease == 0) {
		      numqueries_ideal *= fallback_percent;
		      if (numqueries_ideal < 1) numqueries_ideal = 1;
		      if (o.debugging) { log_write(LOG_STDOUT, " ,     %d\n", (int) numqueries_ideal);
		      windowdecrease++;
		      if (scantype == UDP_SCAN) usleep(250000);
		      }
		    } else if (o.debugging > 1) {
		      log_write(LOG_STDOUT, " ,    .\n");
		    }
		  }
		}
		if (current->state != newstate) {
		  changed++;
		}
		if (current->state != PORT_OPEN &&
		    current->state != PORT_CLOSED) {	
		  numqueries_outstanding--;
		}
		if (current->state == PORT_TESTING && current == testinglist)
		  testinglist = (current->next >= 0)?  &scan[current->next] : NULL;
		else if (current->state == PORT_OPEN && current == openlist)
		  openlist = (current->next >= 0)? &scan[current->next] : NULL;
		if (current->next >= 0) scan[current->next].prev = current->prev;
		if (current->prev >= 0) scan[current->prev].next = current->next;
		current->next = current->prev = -1;
		current->state = newstate;
		addport(&target->ports, current->portno,
			(scantype == UDP_SCAN)? IPPROTO_UDP : IPPROTO_TCP,
			NULL, current->state);	
	      }
	    }
	  }
      }

    testinglist = openlist;
    for(current = openlist; current; current = (current->next >= 0)? &scan[current->next] : NULL) {
      current->state = PORT_FRESH;
      current->trynum = 0;
      if (o.debugging) {
	log_write(LOG_STDOUT, "   ,   %lu   .\n", current->portno);
      }
    }

    openlist = NULL;
    numqueries_ideal = initial_packet_width;
    if (o.debugging)
      log_write(LOG_STDOUT, "   %d .\n", tries);
    if (scantype == UDP_SCAN && changed && (tries + 1) < 100) {
      if (o.debugging) {
	log_write(LOG_STDOUT, " 1/2  -    ICMP-.\n");
      }
      usleep(500000);
    }
  } while(changed && ++tries < 100);

  openlist = testinglist;

  if (o.debugging || o.verbose)
    log_write(LOG_STDOUT, " UDP  FIN/NULL/XMAS  %ld %s  %d .\n",
	    (long) time(NULL) - starttime, (((long) time(NULL) - starttime) == 1)? "" : "",  o.numports);

  for (current = openlist; current;  current = (current->next >= 0)? &scan[current->next] : NULL) {
    if (scantype != UDP_SCAN)
      addport(&target->ports, current->portno, IPPROTO_TCP, NULL, PORT_OPEN);
    else
      addport(&target->ports, current->portno, IPPROTO_UDP, NULL, PORT_OPEN);
  }

 superscan_timedout:

  free(scan);
  close(rawsd);
  pcap_close(pd);
  if (o.numports > 25) {
    if (scantype == UDP_SCAN) {
      if (target->ports.state_counts_udp[PORT_OPEN] == o.numports) {
	if (o.verbose) {
	  error("(      udp- -- ,    )\n");
	}
	for(portno = 0; portno < 65536; portno++)
	  {
	    current_port_tmp = lookupport(&target->ports, portno, IPPROTO_UDP);
	    if (current_port_tmp) {
	      assert(current_port_tmp->state == PORT_OPEN);
	      current_port_tmp->state = PORT_FIREWALLED;
	      target->ports.state_counts[PORT_OPEN]--;
	      target->ports.state_counts[PORT_FIREWALLED]++;
	      target->ports.state_counts_udp[PORT_OPEN]--;
	      target->ports.state_counts_udp[PORT_FIREWALLED]++;
	    }
	  }
      }
    } else {
      if (target->ports.state_counts_tcp[PORT_OPEN] == o.numports) {
	if (o.verbose) {
	  error("(      tcp- -- ,    )\n");
	}
	for(portno = 0; portno < 65536; portno++)
	  {
	    current_port_tmp = lookupport(&target->ports, portno, IPPROTO_TCP);
	    if (current_port_tmp) {
	      assert(current_port_tmp->state == PORT_OPEN);
	      current_port_tmp->state = PORT_FIREWALLED;
	      target->ports.state_counts[PORT_OPEN]--;
	      target->ports.state_counts[PORT_FIREWALLED]++;
	      target->ports.state_counts_tcp[PORT_OPEN]--;
	      target->ports.state_counts_tcp[PORT_FIREWALLED]++;
	    }
	  }
      }
    }
  }
  return;
}

int check_firewallmode(struct hoststruct *target, struct scanstats *ss) {
  struct firewallmodeinfo *fm = &(target->firewallmode);
  struct timeval current_time;
  static struct timeval last_adjust;
  static int init = 0;

  if (!init) {
    gettimeofday(&last_adjust, NULL);
    init = 1;
  }

  if (fm->nonresponsive_ports > 50 && ((double)fm->responsive_ports / (fm->responsive_ports + fm->nonresponsive_ports)) < 0.05) {
    if (fm->active == 0 && o.debugging)
      error("     %s (%s)   .\n", target->name, inet_ntoa(target->host));
    fm->active = 1;
  }

  if (fm->active) {
    gettimeofday(&current_time, NULL);
    if (TIMEVAL_SEC_SUBTRACT(current_time, last_adjust) > 5) {
      ss->numqueries_ideal = MIN(ss->numqueries_ideal + (ss->packet_incr/ss->numqueries_ideal), ss->max_width);
      if (o.debugging) {
	error("     %10.7g   .\n", ss->numqueries_ideal);
      }
      last_adjust = current_time;
    }
  }
  return fm->active;
}

void pos_scan(struct hoststruct *target, unsigned short *portarray, stype scantype) {
  int initial_packet_width;
  struct scanstats ss;
  int rawsd;
  char myname[513];
  int scanflags = 0;
  int victim;
  int senddelay = 0;
  pcap_t *pd = NULL;
  struct bpf_program fcode;
  char err0r[PCAP_ERRBUF_SIZE];
  char filter[512];
  char *p;
  unsigned int ack_number = 0;
  int tries = 0;
  int  res;
  int connecterror = 0;
  unsigned int localnet, netmask;
  int starttime;
  struct hostent *myhostent = NULL;
  struct sockaddr_in sock;
  struct portinfo *scan = NULL,  *current, *next;
  struct portinfolist pil;
  int portlookup[65536];
  struct timeval now;
  struct connectsockinfo csi;
  struct rpcscaninfo rsi;
  unsigned long sequences[3];
  int i;

  if (target->timedout)
    return;

  if (o.debugging)
    log_write(LOG_STDOUT, " pos_scan.\n");

  if (scantype == RPC_SCAN) initial_packet_width = 2;
  else initial_packet_width = 10;

  ss.packet_incr = 4;
  ss.fallback_percent = 0.7;
  ss.numqueries_outstanding = 0;
  ss.ports_left = o.numports;
  ss.alreadydecreasedqueries = 0;

  bzero(&pil, sizeof(pil));

  FD_ZERO(&csi.fds_read);
  FD_ZERO(&csi.fds_write);
  FD_ZERO(&csi.fds_except);
  csi.maxsd = 0;

  target->firewallmode.active = 0;
  target->firewallmode.nonresponsive_ports = 0;
  target->firewallmode.responsive_ports = 0;

  if (o.max_parallelism) {
    ss.max_width = o.max_parallelism;
  } else {
    if (scantype == CONNECT_SCAN) {
      ss.max_width = MAX(5, max_sd() - 4);
    } else {
      ss.max_width = 150;
    }
  }

  if (initial_packet_width > ss.max_width)
    initial_packet_width = ss.max_width;
  ss.numqueries_ideal = initial_packet_width;

  memset(portlookup, 255, sizeof(portlookup));
  bzero(csi.socklookup, sizeof(csi.socklookup));

  if (scantype != RPC_SCAN) {
    scan = safe_malloc(o.numports * sizeof(struct portinfo));
    for(i = 0; i < o.numports; i++) {
      scan[i].state = PORT_FRESH;
      scan[i].portno = portarray[i];
      scan[i].trynum = 0;
      scan[i].prev = i-1;
      scan[i].sd[0] = scan[i].sd[1] = scan[i].sd[2] = -1;
      if (i < o.numports -1) scan[i].next = i+1;
      else scan[i].next = -1;
      portlookup[portarray[i]] = i;
    }
    current = pil.testinglist = &scan[0];
  }

  if ((scantype == SYN_SCAN) || (scantype == WINDOW_SCAN) ||
      (scantype == ACK_SCAN)) {
    if ((rawsd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
      pfatal("   pos_scan");

    broadcast_socket(rawsd);

    get_random_bytes(sequences, sizeof(sequences));


    if (!target->source_ip.s_addr) {
      if (gethostname(myname, MAXHOSTNAMELEN) != 0 &&
	  !((myhostent = gethostbyname(myname))))
	fatal("   !    -S <IP- >  -e < >\n");
      memcpy(&target->source_ip, myhostent->h_addr_list[0], sizeof(struct in_addr));
      if (o.debugging || o.verbose)
	log_write(LOG_STDOUT, "  IP-: %s\n",
		inet_ntoa(target->source_ip));
    }

    pd = my_pcap_open_live(target->device, 100,  (o.spoofsource)? 1 : 0, 20);

    if (pcap_lookupnet(target->device, &localnet, &netmask, err0r) < 0)
      fatal("   : %s\n", err0r);
    p = strdup(inet_ntoa(target->host));
    snprintf(filter, sizeof(filter), "(icmp and dst host %s) or (tcp and src host %s and dst host %s)", inet_ntoa(target->source_ip), p, inet_ntoa(target->source_ip));
    free(p);

    if (islocalhost(&(target->host)))
      filter[0] = '\0';

    if (o.debugging)
      log_write(LOG_STDOUT, "   \n%s\n", filter);
    if (pcap_compile(pd, &fcode, filter, 0, netmask) < 0)
      fatal("    \n%s\n", pcap_geterr(pd));
    if (pcap_setfilter(pd, &fcode) < 0)
      fatal("    \n%s\n", pcap_geterr(pd));
    if (scantype == SYN_SCAN)
      scanflags = TH_SYN;
    else
      scanflags = TH_ACK;
  } else if (scantype == CONNECT_SCAN) {
    rawsd = -1;
    bzero((char *)&sock,sizeof(struct sockaddr_in));
    sock.sin_addr.s_addr = target->host.s_addr;
    sock.sin_family=AF_INET;
  } else {
    get_rpc_procs(&(rsi.rpc_progs), &(rsi.rpc_number));
    scan = safe_malloc(rsi.rpc_number * sizeof(struct portinfo));
    for(i = 0; i < rsi.rpc_number; i++) {
      scan[i].state = PORT_FRESH;
      scan[i].portno = rsi.rpc_progs[i];
      scan[i].trynum = 0;
      scan[i].prev = i-1;
      scan[i].sd[0] = scan[i].sd[1] = scan[i].sd[2] = -1;
      if (i < rsi.rpc_number -1) scan[i].next = i+1;
      else scan[i].next = -1;
    }
    current = pil.testinglist = &scan[0];
    rawsd = -1;
    rsi.rpc_current_port = NULL;
  }

  starttime = time(NULL);

  if (o.debugging || o.verbose) {
    if (scantype == SYN_SCAN)
      log_write(LOG_STDOUT, " SYN-  %s (%s)\n", target->name, inet_ntoa(target->host));
    else if (scantype == CONNECT_SCAN)
      log_write(LOG_STDOUT, " TCP connect()-  %s (%s)\n",target->name, inet_ntoa(target->host));
    else if (scantype == WINDOW_SCAN) {
      log_write(LOG_STDOUT, " Window-  %s (%s)\n",target->name, inet_ntoa(target->host));
      ack_number = get_random_uint();
    }
    else if (scantype == ACK_SCAN) {
      log_write(LOG_STDOUT, " ACK-  %s (%s)\n",target->name, inet_ntoa(target->host));
      ack_number = get_random_uint();
    }
    else {
      log_write(LOG_STDOUT, " RPC-  %s (%s)\n",target->name, inet_ntoa(target->host));
    }
  }

  do {
    ss.changed = 0;
    if (tries > 3 && tries < 10) {
      senddelay += 10000 * (tries - 3);
      if (o.verbose) log_write(LOG_STDOUT, "        %d  (  %d )\n", 10000 * (tries - 3), senddelay);
    } else if (tries >= 10) {
      senddelay += 75000;
      if (o.verbose) log_write(LOG_STDOUT, "        75000  (  %d )\n", senddelay);
    }

    if (senddelay > 200000) {
      ss.max_width = MIN(ss.max_width, 5);
    }

    if (target->timedout)
      goto posscan_timedout;

    if (scantype == RPC_SCAN) {
      rsi.rpc_current_port = nextport(&target->ports, rsi.rpc_current_port,
				      0, PORT_OPEN);
      if (!rsi.rpc_current_port) break;
      pil.testinglist = &scan[0];
      rsi.valid_responses_this_port = 0;
      rsi.rpc_status = RPC_STATUS_UNKNOWN;
    }

    while(pil.testinglist != NULL)
      {

	gettimeofday(&now, NULL);

	if (o.host_timeout && (TIMEVAL_MSEC_SUBTRACT(now, target->host_timeout) >= 0))
	  {
	    target->timedout = 1;
	    goto posscan_timedout;
	  }


	check_firewallmode(target, &ss);

	for(current = pil.testinglist; current ; current = next) {
	  next = (current->next > -1)? &scan[current->next] : NULL;
	  if (current->state == PORT_TESTING) {
	    if (TIMEVAL_SUBTRACT(now, current->sent[current->trynum]) > target->to.timeout) {
	      if (current->trynum > 1 ||
		  (current->trynum > 0 && target->firewallmode.active)) {
		if (scantype == RPC_SCAN) {
		  if (rsi.valid_responses_this_port == 0) {	
		    if (o.debugging) {
		      log_write(LOG_STDOUT, "\n    RPC- .\n  : %hi : %d\n", rsi.rpc_current_port->portno,  rsi.rpc_current_port->proto);
		    }
		    rsi.rpc_status = RPC_STATUS_NOT_RPC;
		    break;
		  }
		  else {
		    target->to.rttvar = MIN(2000000, target->to.rttvar * 1.2);
		  }	
		}
		if (o.debugging) { log_write(LOG_STDOUT, "   %lu     .\n", current->portno); }
		target->firewallmode.nonresponsive_ports++;
		current->state = PORT_FIREWALLED;
		if (current->next > -1) scan[current->next].prev = current->prev;
		if (current->prev > -1) scan[current->prev].next = current->next;
		if (current == pil.testinglist)
		  pil.testinglist = (current->next >= 0)?  &scan[current->next] : NULL;
		current->next = -1;
		current->prev = -1;
		if (scantype != RPC_SCAN) {	
		  if (!pil.firewalled) pil.firewalled = current;
		  else {
		    current->next = pil.firewalled - scan;
		    pil.firewalled = current;
		    scan[current->next].prev = current - scan;	
		  }
		}
		if (scantype == CONNECT_SCAN) {
		  for(i=0; i <= current->trynum; i++) {
		    if (current->sd[i] >= 0) {
		      csi.socklookup[current->sd[i]] = NULL;
		      FD_CLR(current->sd[i], &csi.fds_read);
		      FD_CLR(current->sd[i], &csi.fds_write);
		      FD_CLR(current->sd[i], &csi.fds_except);
		      close(current->sd[i]);
		      current->sd[i] = -1;
		      ss.numqueries_outstanding--;
		    }
		  }
		} else { ss.numqueries_outstanding--; }
	      } else {
		if (o.scan_delay) enforce_scan_delay(NULL);
		if (o.debugging > 1) { log_write(LOG_STDOUT, "    / %lu .\n", current->portno); }
		current->trynum++;
		gettimeofday(&current->sent[current->trynum], NULL);
		now = current->sent[current->trynum];
		if ((scantype == SYN_SCAN) || (scantype == WINDOW_SCAN) || (scantype == ACK_SCAN)) {	
		  if (o.fragscan)
		    send_small_fragz_decoys(rawsd, &target->host, sequences[current->trynum], o.magic_port + tries * 3 + current->trynum, current->portno, scanflags);
		  else
		    send_tcp_raw_decoys(rawsd, &target->host, o.magic_port +
					tries * 3 + current->trynum,
					current->portno,
					sequences[current->trynum],
					ack_number, scanflags, 0, NULL, 0, 0,
					0);

		} else if (scantype == RPC_SCAN) {
		  if (send_rpc_query(&target->host, rsi.rpc_current_port->portno,
				     rsi.rpc_current_port->proto,
				     current->portno, current - scan,
				     current->trynum) == -1) {
		    rsi.rpc_status = RPC_STATUS_NOT_RPC;
		    break;
		  }
		} else {
		  if (ss.numqueries_outstanding >= ss.max_width) {		
		    victim = -1;
		    for(i=0; i < current->trynum; i++)
		      if (current->sd[i] >= 0) {
			victim = i;
			break;
		      }
		    if (victim == -1)
		      fatal("  pos_scan -- ,      alex@nmap.ru\n");
		    csi.socklookup[current->sd[victim]] = NULL;
		    FD_CLR(current->sd[victim], &csi.fds_read);
		    FD_CLR(current->sd[victim], &csi.fds_write);
		    FD_CLR(current->sd[victim], &csi.fds_except);
		    close(current->sd[victim]);
		    current->sd[victim] = -1;
		  } else {
		    ss.numqueries_outstanding++;
		  }
		  res = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		  if (res == -1) pfatal("   pos_scan 143\n");
		  csi.socklookup[res] = current;
		  unblock_socket(res);
		  init_socket(res);
		  sock.sin_port = htons(current->portno);
		  current->sd[current->trynum] = res;		
		  res =  connect(res,(struct sockaddr *)&sock,sizeof(struct sockaddr));
		  if (res != -1) {
		    posportupdate(target, current, current->trynum, scan, &ss, scantype, PORT_OPEN, &pil, &csi);
		  } else {
		    switch(errno) {
		    case EINPROGRESS:
		    case EAGAIN:
		      if (csi.maxsd < current->sd[current->trynum])
			csi.maxsd = current->sd[current->trynum];
		      FD_SET(current->sd[current->trynum], &csi.fds_write);
		      FD_SET(current->sd[current->trynum], &csi.fds_read);
		      FD_SET(current->sd[current->trynum], &csi.fds_except);
		      break;
		    default:
		      if (!connecterror) {	
			connecterror++;
			fprintf(stderr, "   connect (%d):", errno);
			fflush(stdout);
			perror("");
		      }
		    case ECONNREFUSED:
		      posportupdate(target, current, current->trynum, scan, &ss, scantype, PORT_CLOSED, &pil, &csi);
		      break;
		    }  		
		  }
		}
		if (senddelay) usleep(senddelay);
	      }
	    }
	  } else {
	    if (current->state != PORT_FRESH)
	      fatal(" . %d\n", current->state);
	    if (ss.numqueries_outstanding >= (int) ss.numqueries_ideal) break;
	    if (o.scan_delay) enforce_scan_delay(NULL);
	    if (o.debugging > 1) log_write(LOG_STDOUT, "   / %lu\n", current->portno);
	    current->state = PORT_TESTING;
	    current->trynum = 0;
	    ss.numqueries_outstanding++;
	    gettimeofday(&current->sent[0], NULL);
	    if ((scantype == SYN_SCAN) || (scantype == WINDOW_SCAN) ||
		(scantype == ACK_SCAN)) {	
	      if (o.fragscan)
		send_small_fragz_decoys(rawsd, &target->host, sequences[current->trynum], o.magic_port + tries * 3, current->portno, scanflags);
	      else
		send_tcp_raw_decoys(rawsd, &target->host,
				    o.magic_port + tries * 3, current->portno,
				    sequences[current->trynum], ack_number,
				    scanflags, 0, NULL, 0, 0, 0);
	    } else if (scantype == RPC_SCAN) {
	      if (send_rpc_query(&target->host, rsi.rpc_current_port->portno,
				 rsi.rpc_current_port->proto, current->portno,
				 current - scan, current->trynum) == -1) {
		rsi.rpc_status = RPC_STATUS_NOT_RPC;
		break;
	      }
	    } else {
	      res = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	      if (res == -1) pfatal("   pos_scan 11234\n");
	      csi.socklookup[res] = current;
	      unblock_socket(res);
	      init_socket(res);
	      sock.sin_port = htons(current->portno);
	      current->sd[current->trynum] = res;		
	      res =  connect(res,(struct sockaddr *)&sock,sizeof(struct sockaddr));
	      if (res != -1) {
		posportupdate(target, current, current->trynum, scan, &ss, scantype, PORT_OPEN, &pil, &csi);
	      } else {
		switch(errno) {
		case EINPROGRESS:
		case EAGAIN:
		  if (csi.maxsd < current->sd[current->trynum])
		    csi.maxsd = current->sd[current->trynum];
		  FD_SET(current->sd[current->trynum], &csi.fds_write);
		  FD_SET(current->sd[current->trynum], &csi.fds_read);
		  FD_SET(current->sd[current->trynum], &csi.fds_except);
		  break;
		default:
		  if (!connecterror) {	
		    connecterror++;
		    fprintf(stderr, "   connect (%d):", errno);
		    fflush(stdout);
		    perror("");
		  }
		case ECONNREFUSED:
		  posportupdate(target, current, current->trynum, scan, &ss, scantype, PORT_CLOSED, &pil, &csi);
		  break;
		}  		
	      }	
	    }
	    if (senddelay) usleep(senddelay);
	  }
	}
	ss.alreadydecreasedqueries = 0;
	if ((scantype == SYN_SCAN) || (scantype == WINDOW_SCAN) || (scantype == ACK_SCAN))
	  get_syn_results(target, scan, &ss, &pil, portlookup, pd, sequences, scantype);
	else if (scantype == RPC_SCAN) {

	  if (rsi.rpc_status == RPC_STATUS_UNKNOWN) {	
	    get_rpc_results(target, scan, &ss, &pil, &rsi);
	  }
	  if (rsi.rpc_status != RPC_STATUS_UNKNOWN)
	    break;
	}
	else {
	  get_connect_results(target, scan, &ss, &pil, portlookup, sequences, &csi);	
	}

	if (target->timedout)
	  goto posscan_timedout;
      }

    if (scantype == RPC_SCAN) {
      rsi.rpc_current_port->rpc_status = rsi.rpc_status;
      if (rsi.rpc_status == RPC_STATUS_GOOD_PROG) {
	rsi.rpc_current_port->rpc_program = rsi.rpc_program;
	rsi.rpc_current_port->rpc_lowver = rsi.rpc_lowver;
	rsi.rpc_current_port->rpc_highver = rsi.rpc_highver;
      }
      for(i = 0; i < rsi.rpc_number; i++) {
	scan[i].state = PORT_FRESH;
	scan[i].trynum = 0;
	scan[i].prev = i-1;
	if (i < rsi.rpc_number -1) scan[i].next = i+1;
	else scan[i].next = -1;
      }
      current = pil.testinglist = &scan[0];
      pil.firewalled = NULL;
      ss.numqueries_outstanding = 0;
      continue;
    }

    if (ss.numqueries_outstanding != 0) {
      fatal(" 4321897:  : %d   : %d\n", ss.ports_left, ss.numqueries_outstanding);
    }
    if (pil.firewalled) {
      if (tries == 0 || ss.changed) {	
	pil.testinglist = pil.firewalled;
	for(current = pil.testinglist; current ;
	     current = (current->next > -1)? &scan[current->next] : NULL) {
	  current->state = PORT_FRESH;
	  current->trynum = 0;
	  current->sd[0] = current->sd[1] = current->sd[2] = -1;
	  if (o.debugging) {
	    log_write(LOG_STDOUT, "  ,   %lu   .\n", current->portno);
	  }
	}
	pil.firewalled = NULL;
      } else {	
	for(current = pil.firewalled; current ;
	     current = (current->next > -1)? &scan[current->next] : NULL) {
	  addport(&target->ports, current->portno, IPPROTO_TCP, NULL, PORT_FIREWALLED);
	}
	pil.testinglist = NULL;
      }
      tries++;
    }
    ss.numqueries_ideal = initial_packet_width;
    if (o.debugging)
      log_write(LOG_STDOUT, "  %d .\n", tries);
  } while(pil.testinglist && tries < 20);

  if (tries == 20) {
    error(":    20 .\n");
  }

  if (o.verbose)
    log_write(LOG_STDOUT, "%s-  %ld %s  %d .\n", (scantype == WINDOW_SCAN) ? "Window" : (scantype == SYN_SCAN)? "SYN" : (scantype == CONNECT_SCAN)? "TCP connect" : (scantype == RPC_SCAN)? "RPC" : "ACK",  (long) time(NULL) - starttime, (((long) time(NULL) - starttime) == 1)? "" : "", o.numports);

 posscan_timedout:

  free(scan);
  if (rawsd >= 0)
    close(rawsd);
  if (pd)
    pcap_close(pd);
  if (scantype == RPC_SCAN)
    close_rpc_query_sockets();
  return;
}

void posportupdate(struct hoststruct *target, struct portinfo *current,
		   int trynum, struct portinfo *scan,
		   struct scanstats *ss ,stype scantype, int newstate,
		   struct portinfolist *pil, struct connectsockinfo *csi) {
  static int tryident = -1;
  static unsigned int lasttarget = 0;
  struct sockaddr_in mysock;
  int sockaddr_in_len = sizeof(SA);
  int i;
  char owner[1024];

  if (tryident == -1 || target->host.s_addr != lasttarget)
    tryident = o.identscan;
  lasttarget = target->host.s_addr;
  owner[0] = '\0';
  if (current->state != PORT_OPEN && current->state != PORT_CLOSED &&
      current->state != PORT_FIREWALLED && current->state != PORT_TESTING) {
    if (o.debugging) error("    %lu  . : %s (%d)\n", current->portno, statenum2str(current->state), current->state);
    return;
  }

  if (trynum > -1) {
    adjust_timeouts(current->sent[trynum], &(target->to));
    target->firewallmode.responsive_ports++;
  }

  if (trynum == 0) {
    ss->numqueries_ideal = MIN(ss->numqueries_ideal + (ss->packet_incr/ss->numqueries_ideal), ss->max_width);
  } else if (trynum != -1) {
    if (!ss->alreadydecreasedqueries) {
      ss->alreadydecreasedqueries = 1;
      ss->numqueries_ideal *= ss->fallback_percent;
      if (target->firewallmode.active)
	ss->numqueries_ideal *= ss->fallback_percent;
      if (ss->numqueries_ideal < 1.0) ss->numqueries_ideal = 1.0;
    }
  }

  if (newstate == PORT_OPEN && scantype == CONNECT_SCAN && tryident) {
    if (getsockname(current->sd[trynum], (SA *) &mysock,
		    &sockaddr_in_len)) {
      pfatal("getsockname");
    }
    if (getidentinfoz(target->host, ntohs(mysock.sin_port), current->portno, owner) == -1)
      tryident = 0;
  }

  switch(current->state) {
  case PORT_OPEN:
    return;
    break;
  case PORT_FRESH:
    fprintf(stderr, "  : %lu.\n", current->portno);
    return;
    break;
  case PORT_CLOSED:
    if (newstate == PORT_CLOSED)
      return;
    ss->changed++;
    current->state = newstate;
    break;
  case PORT_TESTING:
    if (newstate != PORT_FIREWALLED)
      ss->changed++;
    if (scantype == SYN_SCAN)
      ss->numqueries_outstanding--;
    else {
      for(i=0; i <= current->trynum; i++)
	if (current->sd[i] > -1) {
	  csi->socklookup[current->sd[i]] = NULL;
	  FD_CLR(current->sd[i], &(csi->fds_read));
	  FD_CLR(current->sd[i], &(csi->fds_write));
	  FD_CLR(current->sd[i], &(csi->fds_except));
	  if (current->sd[i] == csi->maxsd)
	    csi->maxsd--;
	  close(current->sd[i]);
	  current->sd[i] = -1;
	  ss->numqueries_outstanding--;
	}
    }
    if (current == pil->testinglist)
      pil->testinglist = (current->next >= 0)? &scan[current->next] : NULL;
    if (current->next >= 0)  scan[current->next].prev = current->prev;
    if (current->prev >= 0)  scan[current->prev].next = current->next;
    break;
  case PORT_FIREWALLED:
    if (newstate != PORT_FIREWALLED)
      ss->changed++;
    if (current == pil->firewalled)
      pil->firewalled = (current->next >= 0)? &scan[current->next] : NULL;
    if (current->next >= 0)  scan[current->next].prev = current->prev;
    if (current->prev >= 0)  scan[current->prev].next = current->next;
    break;
  default:
    fatal("  : %d\n", current->state);
    break;
  }
  current->state = newstate;
  current->next = -1;
  current->prev = -1;
  if (newstate == PORT_OPEN && o.verbose) {
    log_write(LOG_STDOUT, " TCP- %lu (: %s).\n", current->portno, statenum2str(current->state));
    log_flush(LOG_STDOUT);
  }
  addport(&target->ports, current->portno, IPPROTO_TCP, owner, newstate);
  return;
}

inline void adjust_timeouts(struct timeval sent, struct timeout_info *to) {
  int delta = 0;
  struct timeval end;
  gettimeofday(&end, NULL);

  if (o.debugging > 1) {
    log_write(LOG_STDOUT, " :\nSmoothed round-trip time (srtt): %d\nRound-trip time variance (rttvar): %d\nTimeout (to) %d\n\n", to->srtt, to->rttvar, to->timeout);
  }
  if (to->srtt == -1 && to->rttvar == -1) {
    to->srtt = TIMEVAL_SUBTRACT(end, sent);
    to->rttvar = MAX(5000, MIN(to->srtt, 2000000));
    to->timeout = to->srtt + (to->rttvar << 2);
  }
  else {
    delta = TIMEVAL_SUBTRACT(end, sent);
    if (delta >= 8000000 || delta < 0) {
      if (o.verbose)
	error("adjust_timeout:    rtt   %lu  ... .\n", delta);
      return;
    }
    delta -= to->srtt;
    if (delta > 1500000 && delta > 3 * to->srtt + 2 * to->rttvar) {
      if (o.debugging) {
	log_write(LOG_STDOUT, "  delta: %d (srtt %d) ... .\n", delta, to->srtt);
      }
      return;
    }
    to->srtt += delta >> 3;
    to->rttvar += (ABS(delta) - to->rttvar) >> 2;
    to->timeout = to->srtt + (to->rttvar << 2);
  }
  if (to->rttvar > 2300000) {
    fprintf(stderr, " rttvar    2.3 ,   2.0\n");
    to->rttvar = 2000000;
  }

  to->timeout = MAX(to->timeout, o.min_rtt_timeout * 1000);
  to->timeout = MIN(to->timeout, o.max_rtt_timeout * 1000);

  if (o.scan_delay)
    to->timeout = MAX(to->timeout, o.scan_delay * 1000);

  if (o.debugging > 1) {
    log_write(LOG_STDOUT, "Delta %d ==> Smoothed round-trip time (srtt): %d RTT variance (rttvar): %d Timeout (to): %d\n", delta, to->srtt, to->rttvar, to->timeout);
  }

  if (to->srtt < 0 || to->rttvar < 0 || to->timeout < 0 || delta < -50000000) {
    fatal("\n  adjust_timeout.\n :\n   (end) = (%d, %d)\n   (sent)=(%d,%d)\n     (delta) = %d\n    (srtt) = %d\n    (rttvar) = %d\n  (to) = %d\n\n", end.tv_sec, end.tv_usec, sent.tv_sec, sent.tv_usec, delta, to->srtt, to->rttvar, to->timeout);
  }
}

void enforce_scan_delay(struct timeval *tv) {
  static int init = -1;
  static struct timeval lastcall;
  struct timeval now;
  int time_diff;

  if (!o.scan_delay) {
    if (tv) gettimeofday(tv, NULL);
    return;
  }

  if (init == -1) {
    gettimeofday(&lastcall, NULL);
    init = 0;
    if (tv)
      memcpy(tv, &lastcall, sizeof(struct timeval));
    return;
  }

  gettimeofday(&now, NULL);
  time_diff = TIMEVAL_MSEC_SUBTRACT(now, lastcall);
  if (time_diff < o.scan_delay) {
    if (o.debugging > 1) {
      printf("  %d   enforce_scan_delay()\n", o.scan_delay - time_diff);
    }
    usleep((o.scan_delay - time_diff) * 1000);
    gettimeofday(&lastcall, NULL);
  } else
    memcpy(&lastcall, &now, sizeof(struct timeval));
  if (tv) {
    memcpy(tv, &lastcall, sizeof(struct timeval));
  }

  return;
}

int get_connect_results(struct hoststruct *target, struct portinfo *scan,
			struct scanstats *ss, struct portinfolist *pil,
			int *portlookup, unsigned long *sequences,
			struct connectsockinfo *csi) {
  fd_set fds_rtmp, fds_wtmp, fds_xtmp;
  int selectres;
  int selectedfound;
  int optval, optlen = sizeof(int);
  struct timeval timeout;
  int i, sd;
  int trynum;
  char buf[2048];
  struct portinfo *current = NULL;
  struct timeval tv;
  int res;
#ifdef LINUX
  struct sockaddr_in sin,sout;
  int sinlen = sizeof(sin);
  int soutlen = sizeof(sout);
#endif

  res = 0;
  do {
    fds_rtmp = csi->fds_read;
    fds_wtmp = csi->fds_write;
    fds_xtmp = csi->fds_except;
    timeout.tv_sec = 0;
    timeout.tv_usec = 20000;
    selectedfound = 0;

    if (o.host_timeout) {	
      gettimeofday(&tv, NULL);
      if (TIMEVAL_MSEC_SUBTRACT(tv, target->host_timeout) >= 0) {
	target->timedout = 1;
	return 0;
      }
    }

    selectres = select(csi->maxsd+1, &fds_rtmp, &fds_wtmp, &fds_xtmp, &timeout);
    for(sd=0; selectedfound < selectres && sd <= csi->maxsd; sd++) {
      current = csi->socklookup[sd];
      if (!current) continue;
      trynum = -1;
      if  (FD_ISSET(sd, &fds_rtmp)  || FD_ISSET(sd, &fds_wtmp) ||
	   FD_ISSET(sd, &fds_xtmp)) {
	for(i=0; i < 3; i++)
	  if (current->sd[i] == sd) {	
	    trynum = i;
	    break;
	  }

	if (o.debugging > 1 && current != NULL)
	  log_write(LOG_STDOUT, "   %lu ( %d)\n", current->portno, trynum);
	if (FD_ISSET(sd, &fds_rtmp)) {
	  if (o.debugging > 1) log_write(LOG_STDOUT, " ");
	  selectedfound++;
	}
	if (FD_ISSET(sd, &fds_wtmp)) {
	  if (o.debugging > 1) log_write(LOG_STDOUT, " ");
	  selectedfound++;
	}
	if (FD_ISSET(sd, &fds_xtmp)) {
	  if (o.debugging > 1) log_write(LOG_STDOUT, " ");
	  selectedfound++;
	}
	if (o.debugging > 1 && current != NULL)
	  log_write(LOG_STDOUT, "\n");

	assert(trynum != -1);

	if (getsockopt(sd, SOL_SOCKET, SO_ERROR, (char *) &optval, &optlen) != 0)
	  optval = errno;

	switch(optval) {
	case 0:
#ifdef LINUX
	  if (!FD_ISSET(sd, &fds_rtmp)) {
	    res = send(current->sd[trynum], "", 0, 0);

	    if (res < 0) {
	      if (o.debugging > 1) {
		log_write(LOG_STDOUT, "   %lu  0- : ", current->portno);
		perror("");
	      }
	      posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_CLOSED, pil, csi);
	    } else {
	      if (getpeername(sd, (struct sockaddr *) &sin, &sinlen) < 0) {
		pfatal("  getpeername  connect_results   %hi\n", current->portno);
	      } else {
		if (current->portno != ntohs(sin.sin_port)) {
		  error("  %hi  .   %hi\n", current->portno, ntohs(sin.sin_port));
		}
	      }

	      if (getsockname(sd, (struct sockaddr *) &sout, &soutlen) < 0) {
		pfatal("  getsockname   %hi\n", current->portno);
	      }
	      if (htons(sout.sin_port) == current->portno) {
		posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_CLOSED, pil, csi);
	      } else {
		posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_OPEN, pil, csi);
	      }
	    }
	  } else {
	    posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_OPEN, pil, csi);
	  }
#else
	  posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_OPEN, pil, csi);
#endif
	  break;
	case ECONNREFUSED:
	  posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_CLOSED, pil, csi);
	  break;
	case EHOSTUNREACH:
	  posportupdate(target, current, trynum, scan, ss, CONNECT_SCAN, PORT_FIREWALLED, pil, csi);
	  break;
	case ENETDOWN:
	case ENETUNREACH:
	case ENETRESET:
	case ECONNABORTED:
	case ETIMEDOUT:
	case EHOSTDOWN:
	  snprintf(buf, sizeof(buf), "  SO_ERROR   %s (%d)\n", inet_ntoa(target->host), optval);
	  perror(buf);
	  return -1;
	  break;
	default:
	  snprintf(buf, sizeof(buf), "     %s (%d)\n", inet_ntoa(target->host), optval);
	  perror(buf);
	  break;
	}
      } else continue;
    }
  } while(ss->numqueries_outstanding > 0 && selectres > 0);

  return 0;
}


void get_syn_results(struct hoststruct *target, struct portinfo *scan,
		     struct scanstats *ss, struct portinfolist *pil,
		     int *portlookup, pcap_t *pd, unsigned long *sequences,
		     stype scantype) {

  struct ip *ip;
  int bytes;
  struct tcphdr *tcp;
  int trynum;
  int newstate = -1;
  int i;
  int newport;
  struct portinfo *current = NULL;
  struct icmp *icmp;
  struct ip *ip2;
  unsigned short *data;
  struct timeval tv;

  while (ss->numqueries_outstanding > 0 &&
	 (ip = (struct ip*) readip_pcap(pd, &bytes, target->to.timeout))) {
    if (bytes < (4 * ip->ip_hl) + 4)
      continue;
    current = NULL;
    trynum = newport = -1;
    newstate = PORT_UNKNOWN;

    if (o.host_timeout) {	
      gettimeofday(&tv, NULL);
      if (TIMEVAL_MSEC_SUBTRACT(tv, target->host_timeout) >= 0) {
	target->timedout = 1;
	return;
      }
    }
    if (ip->ip_src.s_addr == target->host.s_addr && ip->ip_p == IPPROTO_TCP) {
      tcp = (struct tcphdr *) (((char *) ip) + 4 * ip->ip_hl);
      i = ntohs(tcp->th_dport);
      if (i < o.magic_port || i > o.magic_port + 15) {
	if (o.debugging > 1)
	  error("  SYN-  TCP-   %d (magic- %d) ... .\n", i, o.magic_port);
	continue;
      }
      newport = ntohs(tcp->th_sport);
      if (ip->ip_src.s_addr == target->source_ip.s_addr && ((tcp->th_flags == TH_ACK) || (tcp->th_flags == TH_SYN))) {
	continue;
      }
      if (portlookup[newport] < 0) {
	if (o.debugging) {
	  log_write(LOG_STDOUT, "      %d:\n", ntohs(tcp->th_sport));
	  readtcppacket((char *)ip, bytes);
	}
	current = NULL;
	continue;
      }	

      current = &scan[portlookup[newport]];
      for(i=0; i < 3; i++) {
	if (MOD_DIFF(sequences[i],ntohl(tcp->th_ack)) < 5)
	  break;
      }
      if (i < 3) trynum = i;
      else {
	if (o.debugging)
	  log_write(LOG_STDOUT, "  ACK  : %lX\n", (unsigned long) ntohl(tcp->th_ack));
	trynum = (current->trynum == 0)? 0 : -1;	
      }
      if (current->trynum < trynum) {
	if (o.debugging) 	
	  error("\n SYN-   trynum  %d ( %hi),\n     trynum  %d.\n,    .\n", trynum, newport, current->trynum);
	trynum = -1;
      }
      if (scantype == SYN_SCAN) {
	if (tcp->th_flags & TH_RST) {
	  newstate = PORT_CLOSED;
	} else if ((tcp->th_flags & (TH_SYN|TH_ACK)) == (TH_SYN|TH_ACK)) {
	  newstate = PORT_OPEN;
	} else {
	  if (o.debugging)
	    error("  SYN-      tcp-: %d\n", tcp->th_flags);
	  continue;
	}
      }
      else if (scantype == WINDOW_SCAN) {
	if (tcp->th_flags & TH_RST) {
	  if (tcp->th_win) {
	    newstate = PORT_OPEN;
	  } else {
	    newstate = PORT_CLOSED;
	  }
	} else {
	  if (o.debugging)
	    error("  Window-      tcp-: %d\n", tcp->th_flags);
	  continue;
	}
      }
      else if (scantype == ACK_SCAN) {
	if (tcp->th_flags & TH_RST) {	
	  newstate = PORT_UNFIREWALLED;
	} else {
	  if (o.debugging)
	    error("  ACK-      tcp-: %d\n", tcp->th_flags);
	  continue;
	}
      } else {
	fatal("  !  get_syn_results   !    alex@nmap.ru\n");
      }
    } else if (ip->ip_p == IPPROTO_ICMP) {
      icmp = (struct icmp *) ((char *)ip + 4 * ip->ip_hl);
      ip2 = (struct ip *) (((char *) ip) + 4 * ip->ip_hl + 8);
      if (bytes <= 4 * ip->ip_hl + 28 ||
	  bytes <= 4 * ip->ip_hl + 8 +
	  4 * ip2->ip_hl + 4)
	{
	  if (o.debugging) {
	    error("ICMP-   (%d )\n", bytes);
	  }
	  continue;
	}

      if (ip2->ip_dst.s_addr != target->host.s_addr) {
	if (o.debugging > 1)
	  error(" ICMP-        .\n");
	continue;
      }

      data = (unsigned short *) ((char *)ip2 + 4 * ip2->ip_hl);
      if (icmp->icmp_type == 3) {
	if (icmp->icmp_code != 1 && icmp->icmp_code != 2 &&
	    icmp->icmp_code != 3 && icmp->icmp_code != 13 &&
	    icmp->icmp_code != 9 && icmp->icmp_code != 10) {
	  error(" / 3/%d ICMP- ' ':", icmp->icmp_code);
	  hdump((unsigned char *)icmp, ntohs(ip->ip_len) - sizeof(struct ip));
	  continue;
	}

	newport = ntohs(data[1]);
	if (portlookup[newport] >= 0) {
	  current = &scan[portlookup[newport]];
	  trynum = (current->trynum == 0)? 0 : -1;
	  newstate = PORT_FIREWALLED;
	} else {
	  if (o.debugging) {
	    error(" / 3/%d ICMP- ' ':",
		  icmp->icmp_code);
	    hdump((unsigned char *)icmp, ntohs(ip->ip_len) - sizeof(struct ip));
	  }
	  continue;
	}	
      }
    }
    if (current) {
      posportupdate(target, current, trynum, scan, ss, SYN_SCAN, newstate,
		    pil, NULL);
      current = NULL;
      trynum = -1;
      newstate = PORT_UNKNOWN;
    }
  }
  return;
}

int ftp_anon_connect(struct ftpinfo *ftp) {
  int sd;
  struct sockaddr_in sock;
  int res;
  char recvbuf[2048];
  char command[512];

  if (o.verbose || o.debugging)
    log_write(LOG_STDOUT, "   ftp://%s:%s@%s:%i\n", ftp->user, ftp->pass,
	    ftp->server_name, ftp->port);

  if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    perror("    ftp_anon_connect\n");
    return 0;
  }

  sock.sin_family = AF_INET;
  sock.sin_addr.s_addr = ftp->server.s_addr;
  sock.sin_port = htons(ftp->port);
  res = connect(sd, (struct sockaddr *) &sock, sizeof(struct sockaddr_in));
  if (res < 0) {
    fprintf(stderr, " ftp proxy-   '  FTP'  !\n");
    exit(1);
  }
  if (o.verbose || o.debugging) log_write(LOG_STDOUT, " :");
  while ((res = recvtime(sd, recvbuf, 2048,7)) > 0)
    if (o.debugging || o.verbose) {
      recvbuf[res] = '\0';
      log_write(LOG_STDOUT, "%s", recvbuf);
    }
  if (res < 0) {
    perror("    ftp-.");
    exit(1);
  }

  snprintf(command, 511, "USER %s\r\n", ftp->user);

  send(sd, command, strlen(command), 0);
  res = recvtime(sd, recvbuf, 2048,12);
  if (res <= 0) {
    perror("    ftp-.");
    exit(1);
  }
  recvbuf[res] = '\0';
  if (o.debugging) log_write(LOG_STDOUT, "  , : %s\n", recvbuf);
  if (recvbuf[0] == '5') {
    fprintf(stderr, " ftp proxy-   '  FTP'    \"%s\"\n",
	    ftp->user);
    exit(1);
  }

  snprintf(command, 511, "PASS %s\r\n", ftp->pass);

  send(sd, command, strlen(command), 0);
  res = recvtime(sd, recvbuf, 2048,12);
  if (res < 0) {
    perror("    ftp-.\n");
    exit(1);
  }
  if (!res) fprintf(stderr, "    ftp-  ...\n");
  else {
    recvbuf[res] = '\0';
    if (o.debugging) log_write(LOG_STDOUT, " , : %s", recvbuf);
    if (recvbuf[0] == '5') {
      fprintf(stderr, "      ftp- (%s/%s)\n",
	      ftp->user, ftp->pass);
      exit(1);
    }
  }
  while ((res = recvtime(sd, recvbuf, 2048,2)) > 0)
    if (o.debugging) {
      recvbuf[res] = '\0';
      log_write(LOG_STDOUT, "%s", recvbuf);
    }
  if (res < 0) {
    perror("    ftp-.\n");
    exit(1);
  }
  if (o.verbose) log_write(LOG_STDOUT, "  ftp-  !\n");

  ftp->sd = sd;
  return sd;
}

int recvtime(int sd, char *buf, int len, int seconds) {

  int res;
  struct timeval timeout;
  fd_set readfd;

  timeout.tv_sec = seconds;
  timeout.tv_usec = 0;
  FD_ZERO(&readfd);
  FD_SET(sd, &readfd);
  res = select(sd + 1, &readfd, NULL, NULL, &timeout);
  if (res > 0) {
    res = recv(sd, buf, len, 0);
    if (res >= 0) return res;
    perror("recv  recvtime");
    return 0;
  }
  else if (!res) return 0;
  perror("select()  recvtime");
  return -1;
}

void bounce_scan(struct hoststruct *target, unsigned short *portarray,
		 struct ftpinfo *ftp) {
  int starttime,  res , sd = ftp->sd,  i=0;
  char *t = (char *)&target->host;
  int retriesleft = FTP_RETRIES;
  char recvbuf[2048];
  char targetstr[20];
  char command[512];
  unsigned short portno,p1,p2;
  struct timeval now;

  snprintf(targetstr, 20, "%d,%d,%d,%d,", UC(t[0]), UC(t[1]), UC(t[2]), UC(t[3]));

  starttime = time(NULL);
  if (o.verbose || o.debugging)
    log_write(LOG_STDOUT, " TCP-  '  FTP'  %s (%s)\n",
	    target->name,  inet_ntoa(target->host));
  for(i=0; portarray[i]; i++) {

    if (o.host_timeout) {
      gettimeofday(&now, NULL);
      if ((TIMEVAL_MSEC_SUBTRACT(now, target->host_timeout) >= 0))
	{
	  target->timedout = 1;
	  return;
	}
    }

    portno = htons(portarray[i]);
    p1 = ((unsigned char *) &portno)[0];
    p2 = ((unsigned char *) &portno)[1];
    snprintf(command, 512, "PORT %s%i,%i\r\n", targetstr, p1,p2);
    if (o.debugging) log_write(LOG_STDOUT, "  : %s\n", command);
    if (send(sd, command, strlen(command), 0) < 0) {
      perror("  bounce_scan");
      if (retriesleft) {
	if (o.verbose || o.debugging)
	  log_write(LOG_STDOUT, "Ftp proxy-  !  .\n");
	retriesleft--;
	close(sd);
	ftp->sd = ftp_anon_connect(ftp);
	if (ftp->sd < 0) return;
	sd = ftp->sd;
	i--;
      }
      else {
	fprintf(stderr, "  ,  .  .\n");
	close(sd);
	ftp->sd = -1;
	return;
      }
    } else {
      res = recvtime(sd, recvbuf, 2048,15);
      if (res <= 0) perror("    ftp-.\n");

      else {
	recvbuf[res] = '\0';
	if (o.debugging) log_write(LOG_STDOUT, "   %i: %s\n",
				 portarray[i],  recvbuf);
	if (recvbuf[0] == '5') {
	  if (portarray[i] > 1023) {
	    fprintf(stderr, "FTP-     .\n");
	    exit(1);
	  }
	  else {
	    fprintf(stderr, "FTP-     ,\n");
	    while(portarray[i] && portarray[i] < 1024) i++;
	    if (!portarray[i]) {
	      fprintf(stderr, "    .  .\n");
	      exit(1);
	    }
	  }
	}
	else
	  if (send(sd, "LIST\r\n", 6, 0) > 0) {
	    res = recvtime(sd, recvbuf, 2048,12);
	    if (res <= 0) {
	      perror("    ftp-.\n");
	    }
	    else {
	      recvbuf[res] = '\0';
	      if (o.debugging) log_write(LOG_STDOUT, "   LIST: %s\n", recvbuf);
	      if (!strncmp(recvbuf, "500", 3)) {
		if (o.verbose || o.debugging)
		  fprintf(stderr, "   FTP- ... .\n");
		res = recvtime(sd, recvbuf, 2048,10);
	      }
	      if (recvbuf[0] == '1' || recvbuf[0] == '2') {
		if (o.verbose || o.debugging) log_write(LOG_STDOUT, " %i  .\n",
						      portarray[i]);
		addport(&target->ports, portarray[i], IPPROTO_TCP, NULL, PORT_OPEN);
		if (recvbuf[0] == '1') {
		  res = recvtime(sd, recvbuf, 2048,5);
		  recvbuf[res] = '\0';
		  if (res > 0) {
		    if (o.debugging) log_write(LOG_STDOUT, " : %s\n", recvbuf);
		    if (recvbuf[0] == '4' && recvbuf[1] == '2' &&
			recvbuf[2] == '6') {	      	
		      deleteport(&target->ports, portarray[i], IPPROTO_TCP);
		      if (o.debugging || o.verbose)
			log_write(LOG_STDOUT, " %i  .\n", portarray[i]);
		    }
		  }
		}
	      } else {
		addport(&target->ports, portarray[i], IPPROTO_TCP, NULL, PORT_CLOSED);
	      }
	    }
	  }
      }
    }
  }

  if (o.debugging || o.verbose)
    log_write(LOG_STDOUT, " '  FTP'  %d   %ld .\n",
	    o.numports, (long) time(NULL) - starttime);
  return;
}
int parse_bounce(struct ftpinfo *ftp, char *url) {
  char *p = url,*q, *s;

  if ((q = strrchr(url, '@'))){
    *(q++) = '\0';
    if ((s = strchr(q, ':')))
      {
	*(s++) = '\0';
	strncpy(ftp->server_name, q, MAXHOSTNAMELEN);
	ftp->port = atoi(s);
      }
    else  strncpy(ftp->server_name, q, MAXHOSTNAMELEN);

    if ((s = strchr(p, ':'))) {
      *(s++) = '\0';
      strncpy(ftp->user, p, 63);
      strncpy(ftp->pass, s, 255);
    }
    else {
      log_write(LOG_STDOUT, "%s    USERNAME,   : %s\n",
	      p, ftp->pass);
      strncpy(ftp->user, p, 63);
    }
  }
  else
    if ((s = strchr(url, ':'))) {
      *(s++) = '\0';
      strncpy(ftp->server_name, url, MAXHOSTNAMELEN);
      ftp->port = atoi(s);
    }
    else
      strncpy(ftp->server_name, url, MAXHOSTNAMELEN);

  ftp->user[63] = ftp->pass[255] = ftp->server_name[MAXHOSTNAMELEN] = 0;

  return 1;
}
char *logtypes[LOG_TYPES]={" "," ","HTML","$cR1pT |<1dd13"};

int log_open(int logt, int append, char *filename)
{
  int i=0;
  if (logt<=0 || logt>LOG_MASK) return -1;
  while ((logt&1)==0) { i++; logt>>=1; }
  if (o.logfd[i]) fatal("    %s    .\n",logtypes[i]);
  if (*filename == '-' && *(filename + 1) == '\0')
    {
      o.logfd[i]=stdout;
      o.nmap_stdout = fopen("/dev/null", "w");
      if (!o.nmap_stdout)
	fatal("  /dev/null  stdout  .\n");
  }
  else
    {
      if (o.append_output)
	o.logfd[i] = fopen(filename, "a");
      else
	o.logfd[i] = fopen(filename, "w");
      if (!o.logfd[i])
	fatal("  %s   %s  .\n", logtypes[i], filename);
    }
  return 1;
}

void skid_output(unsigned char *s)
{
  int i;
  for (i=0;s[i];i++)
    if (rand()%2==0)
      switch(s[i])
	{
	case 'A': s[i]='4'; break;
	case 'e':
	case 'E': s[i]='3'; break;
	case 'i':
	case 'I': s[i]="!|1"[rand()%3]; break;
	case 'o':
	case 'O': s[i]='0'; break;
	case 's':
	case 'S':
	  if (s[i+1] && !isalnum(s[i+1]))
	    s[i] = 'z';
	  else s[i] = '$';
	  break;
	case 'z': s[i]='s'; break;
	case 'Z': s[i]='S'; break;
	}
}

void log_write(int logt, char *fmt, ...)
{
  va_list  ap;
  int i,l=logt,skid=1;
  char buffer[1000];

  va_start(ap, fmt);
  if (l & LOG_STDOUT) {
    vfprintf(o.nmap_stdout, fmt, ap);
    l-=LOG_STDOUT;
  }
  if (l & LOG_SKID_NOXLT) { skid=0; l -= LOG_SKID_NOXLT; l |= LOG_SKID; }
  if (l<0 || l>LOG_MASK) return;
  for (i=0;l;l>>=1,i++)
    {
      if (!o.logfd[i] || !(l&1)) continue;
      vsnprintf(buffer,sizeof(buffer)-1,fmt,ap);
      if (skid && ((1<<i)&LOG_SKID)) skid_output(buffer);
      fwrite(buffer,1,strlen(buffer),o.logfd[i]);
    }
  va_end(ap);
}

void log_close(int logt)
{
  int i;
  if (logt<0 || logt>LOG_MASK) return;
  for (i=0;logt;logt>>=1,i++) if (o.logfd[i] && (logt&1)) fclose(o.logfd[i]);
}

void log_flush(int logt) {
  int i;

  if (logt & LOG_STDOUT) {
    fflush(o.nmap_stdout);
    logt -= LOG_STDOUT;
  }
  if (logt & LOG_SKID_NOXLT)
    fatal("   log_flush()  LOG_SKID_NOXLT\n");

  if (logt<0 || logt>LOG_MASK) return;

  for (i=0;logt;logt>>=1,i++)
    {
      if (!o.logfd[i] || !(logt&1)) continue;
      fflush(o.logfd[i]);
    }

}

void log_flush_all() {
  int fileno;

  for(fileno = 0; fileno < LOG_TYPES; fileno++) {
    if (o.logfd[fileno]) fflush(o.logfd[fileno]);
  }
  fflush(stdout);
  fflush(stderr);
}

void reaper(int signo) {
  int status;
  pid_t pid;

  if ((pid = wait(&status)) == -1) {
    gh_perror("  .\n");
  } else {
    fprintf(stderr, "\n[%d  =%d (%s)]\nnmap> ", (int) pid, status, (status == 0)? ""  : "");
  }
}

void sigdie(int signo) {
  switch(signo) {
  case SIGINT:
    fprintf(stderr, "  SIGINT, .\n");
    break;
  case SIGTERM:
    fprintf(stderr, "  SIGTERM, .\n");
    break;
  case SIGHUP:
    fprintf(stderr, "  SIGHUP, .\n");
    break;
  case SIGSEGV:
    fprintf(stderr, "  SIGSEGV, .\n");
    if (o.debugging) abort();
    break;
  case SIGBUS:
    fprintf(stderr, "  SIGBUS, .\n");
    break;
  default:
    fprintf(stderr, "  %d, .\n", signo);
    break;
  }
  fflush(stdout);
  log_close(LOG_MACHINE|LOG_NORMAL|LOG_SKID);
  exit(1);
}

int nmap_fetchfile(char *filename_returned, int bufferlen, char *file) {
  char *dirptr;
  int res;
  int foundsomething = 0;
  struct passwd *pw;
  char dot_buffer[512];
  static int warningcount = 0;

  if ((dirptr = getenv("NMAPDIR"))) {
    res = snprintf(filename_returned, bufferlen, "%s/%s", dirptr, file);
    if (res > 0 && res < bufferlen) {
      if (fileexistsandisreadable(filename_returned))
	foundsomething = 1;
    }
  }
  if (!foundsomething) {
    pw = getpwuid(getuid());
    if (pw) {
      res = snprintf(filename_returned, bufferlen, "%s/.runmap/%s", pw->pw_dir, file);
      if (res > 0 && res < bufferlen) {
	if (fileexistsandisreadable(filename_returned))
	  foundsomething = 1;
      }
    }
    if (!foundsomething && getuid() != geteuid()) {
      pw = getpwuid(geteuid());
      if (pw) {
	res = snprintf(filename_returned, bufferlen, "%s/runmap/%s", pw->pw_dir, file);
	if (res > 0 && res < bufferlen) {
	  if (fileexistsandisreadable(filename_returned))
	    foundsomething = 1;
	}
      }
    }
  }
  if (!foundsomething) {
    res = snprintf(filename_returned, bufferlen, "%s/%s", NMAPDATADIR, file);
    if (res > 0 && res < bufferlen) {
      if (fileexistsandisreadable(filename_returned))
	foundsomething = 1;
    }
  }
  if (foundsomething && (*filename_returned != '.')) {
    res = snprintf(dot_buffer, sizeof(dot_buffer), "./%s", file);
    if (res > 0 && res < bufferlen) {
      if (fileexistsandisreadable(dot_buffer)) {
	if (warningcount++ < 5)
	  error("\n!   %s  %s     .\n     %s.\n NMAPDIR=.        .\n", filename_returned, dot_buffer, filename_returned);
      }
    }
  }

  if (!foundsomething) {
    res = snprintf(filename_returned, bufferlen, "./%s", file);
    if (res > 0 && res < bufferlen) {
      if (fileexistsandisreadable(filename_returned))
	foundsomething = 1;
    }
  }

  if (!foundsomething) {
    filename_returned[0] = '\0';
    return -1;
  }

  if (o.debugging > 1)
    error("  : %s\n", filename_returned);

  return 0;

}

int fileexistsandisreadable(char *pathname) {
  FILE *fp;
  fp = fopen(pathname, "r");
  if (fp) fclose(fp);
  return (fp == NULL)? 0 : 1;
}

void output_rangelist_given_ports_to_machine_output(unsigned short *ports,
						    int numports) {
int i, previous_port = -2, range_start = -2, port;
char outpbuf[128];

 for(i=0; i <= numports; i++) {
   port = (i < numports)? ports[i] : 0xABCDE;
   if (port != previous_port + 1) {
     outpbuf[0] = '\0';
     if (range_start != previous_port && range_start != -2)
       sprintf(outpbuf, "-%hi", previous_port);
     if (port != 0xABCDE) {
       if (range_start != -2)
	 strcat(outpbuf, ",");
       sprintf(outpbuf + strlen(outpbuf), "%hi", port);
     }
     log_write(LOG_MACHINE, "%s", outpbuf);
     range_start = port;
   }
   previous_port = port;
 }
}

void output_ports_to_machine_parseable_output(unsigned short *ports,
					      int numports, int tcpscan,
					      int udpscan) {
  int tcpportsscanned = (tcpscan)? numports : 0;
  int udpportsscanned = (udpscan)? numports : 0;
 log_write(LOG_MACHINE, " : TCP(%d;", tcpportsscanned);
 if (tcpportsscanned)
   output_rangelist_given_ports_to_machine_output(ports, tcpportsscanned);
 log_write(LOG_MACHINE, ") UDP(%d;", udpportsscanned);
 if (udpportsscanned)
   output_rangelist_given_ports_to_machine_output(ports, udpportsscanned);
 log_write(LOG_MACHINE, ")\n");
}